"""
test planar diagrams
"""
#pylint:disable=import-error,invalid-name,too-many-locals
from typing import cast,List,Tuple
from sympy import symbols, Symbol, Expr, Add, Mul, Pow, Integer, Rational, UnevaluatedExpr
from plabic import PlanarNetwork, determinant, ChipType

def test_planar() -> None:
    """
    example is taken from https://arxiv.org/pdf/math/9912128.pdf figure 1
    """

    def annotated_symbols(*args, **kwargs) -> Tuple[Symbol, ...]:
        """
        because symbols is typed Any
        despite the fact that for the arguments we are using it returns an tuple of Symbols
        """
        return symbols(*args, **kwargs)
    A, B, C, D, E, F, G, H, I = annotated_symbols(
        'a,b,c,d,e,f,g,h,i', commutative=True)
    ZERO, ONE = annotated_symbols('zero,one', commutative=True)
    p = PlanarNetwork(3, edge_list=[[(3, 2, A), (3, 3, ONE)], [(3, 2, C), (2, 1, B)], [
                      (2, 2, E), (1, 1, D)], [(3, 3, F), (2, 3, G), (1, 2, H)], [(2, 3, I)]],
                      multiplicative_identity=ONE,
                      additive_identity=ZERO)
    expected_weight_matrix: List[List[Expr]] = [
        [D, D*H, D*H*I], [B*D, B*D*H+E, B*D*H*I+E*(G+I)],\
            [A*B*D, A*B*D*H+(A+C)*E, A*B*D*H*I+(A+C)*E*(G+I)+F]]
    for i in range(1, 4):
        for j in range(1, 4):
            w_ij = cast(Expr, p.weight_matrix(i, j))
            w_ij = w_ij.subs({ZERO: 0.0, ONE: 1.0})
            assert w_ij.equals(expected_weight_matrix[i-1][j-1])
    expected_23_vertex_disjoint_weights = [B*C*D*E*G*H, E*F, B*D*F*H]
    for idx,weight in enumerate(p.vertex_disjoint_collection({2, 3}, {2, 3})):
        weight = cast(Expr, weight)
        weight = weight.subs({ZERO: 0.0, ONE: 1.0})
        assert weight.equals(expected_23_vertex_disjoint_weights[idx])
    delta_23_23 = cast(Expr, p.lindstrom_minor({2, 3}, {2, 3}))
    delta_23_23 = delta_23_23.subs({ZERO: 0.0, ONE: 1.0})
    expected = (B*C*D*E*G*H + B*D*F*H + E*F)*1.0
    assert delta_23_23 == expected
    slice_23 = (2, 3)
    my_minor = [[expected_weight_matrix[cur_i-1][cur_j-1]
                 for cur_i in slice_23] for cur_j in slice_23]
    expected_2 = determinant(my_minor, ONE,
                             lambda z: z.subs(
                                 {ZERO: 0.0, ONE: 1.0}).simplify(),
                             ZERO, -ONE)
    assert expected_2 == expected

    delta_123_123 = cast(Expr, p.lindstrom_minor({2, 3, 1}, {1, 2, 3}))
    delta_123_123 = delta_123_123.subs({ZERO: 0.0, ONE: 1.0})
    expected_determinant = determinant(expected_weight_matrix,
                                       ONE,
                                       lambda z: z.subs(
                                           {ZERO: 0.0, ONE: 1.0}).simplify(),
                                       ZERO,
                                       -ONE)
    assert delta_123_123 == expected_determinant
    def nn_with_my_symbols(edge_weight : Expr) -> bool:
        """
        a check that edge weight is nonnegative
        provided that the given symbols are nonnnegative
        checks with subtraction free
        """
        if isinstance(edge_weight,(Integer,Rational)):
            return edge_weight>0
        if isinstance(edge_weight,Symbol):
            return edge_weight in {ONE, A, B, C, D, E, F, G, H, I}
        op = edge_weight.func
        args = edge_weight.args
        if op == UnevaluatedExpr:
            return nn_with_my_symbols(cast(Expr,args[0]))
        if op in [Add,Mul]:
            return all(nn_with_my_symbols(cast(Expr,arg)) for arg in args)
        if op == Pow:
            return nn_with_my_symbols(cast(Expr,args[0])) and isinstance(args[1],(Integer,Rational))
        return False

    tnn = p.totally_nonnegative(nn_with_my_symbols)
    assert tnn
    assert p.chip_type == [(ChipType.DOWN,2),(ChipType.FLAT,3),\
                           (ChipType.DOWN,2),(ChipType.DOWN,1),\
                           (ChipType.FLAT,2),(ChipType.FLAT,1),\
                           (ChipType.FLAT,3),(ChipType.UP,3),\
                           (ChipType.UP,2),(ChipType.UP,3)]



"""
double wiring diagram
also include single wiring diagram
"""

from functools import reduce
from typing import Tuple,List,Dict,Iterator,Set,Any,cast
#pylint:disable=import-error
from .plabic_diagram import PlabicGraph, BiColor

Point = Tuple[float,float]

class WiringDiagram:
    """
    a single or double wiring diagram
    """
    def __init__(self,my_n : int, my_word : List[int]):
        if my_n<=1:
            msg = "The number of strands must be natural number greater than 1"
            raise ValueError(msg)
        def valid_letter(letter : int) -> bool:
            """
            is (letter,letter+1) a transposition for S_{my_n}
            """
            if letter<0:
                letter = - letter
            return 0<letter<my_n

        if not all(valid_letter(z) for z in my_word):
            raise ValueError(f"The letters in the double word must be 1 to {my_n} or negative that")
        my_sets = (set([abs(letter),abs(letter)+1]) for letter in my_word)
        all_wires_touched = reduce(lambda acc,new : acc.union(new),my_sets,cast(Set[int],set()))
        if len(all_wires_touched) < my_n:
            raise ValueError("There were some indices that never were involved in a swap")
        self.my_word = my_word
        self.my_n = my_n
        self.num_positive_letters = sum(1 if z>0 else 0 for z in my_word)
        self.num_negative_letters = sum(1 if z<0 else 0 for z in my_word)
        self.is_positive_order_flip = False
        self.is_negative_order_flip = False
        for _ in self.chamber_minors():
            pass

    def chamber_minors(self) -> Iterator[Tuple[Set[int],Set[int]]]:
        """
        the chamber minors of a double wiring diagram
        are the labels of the lines below each face
        here they are given as the face just to the left of
        each letter in my_word
        and then all the leftovers on the very
        right hand side of the diagram
        """
        positive_lines = list(range(1,self.my_n+1))
        negative_lines = list(range(1,self.my_n+1))
        negative_lines.reverse()
        for cur_word in self.my_word:
            abs_cur_word = cur_word if cur_word>0 else -cur_word
            flip_positive_lines = cur_word>0
            positive_lines_below = positive_lines[0:abs_cur_word]
            negative_lines_below = negative_lines[0:abs_cur_word]
            yield set(positive_lines_below),set(negative_lines_below)
            if flip_positive_lines:
                positive_lines[abs_cur_word-1],positive_lines[abs_cur_word] = \
                    positive_lines[abs_cur_word],positive_lines[abs_cur_word-1]
            else:
                negative_lines[abs_cur_word],negative_lines[abs_cur_word-1] = \
                    negative_lines[abs_cur_word-1],negative_lines[abs_cur_word]
        self.is_positive_order_flip = positive_lines_below == list(range(self.my_n,0,-1))
        self.is_negative_order_flip = negative_lines_below == list(range(self.my_n,0,-1))
        for pretend_letter in range(1,self.my_n+1):
            positive_lines_below = positive_lines[0:pretend_letter]
            negative_lines_below = negative_lines[0:pretend_letter]
            yield set(positive_lines_below),set(negative_lines_below)

    # pylint:disable=too-many-locals,too-many-branches,too-many-statements
    def to_plabic(self) -> PlabicGraph:
        """
        a plabic graph with bridges for each letter
        """
        last_on_this_strand = [[0]*self.my_n]
        for letter in self.my_word:
            if letter<0:
                letter = - letter
            cur_last_on_this_strand = last_on_this_strand[-1].copy()
            cur_last_on_this_strand[letter-1] += 1
            cur_last_on_this_strand[letter] += 1
            last_on_this_strand.append(cur_last_on_this_strand)
        very_last_numbers = last_on_this_strand[-1]
        my_init_data : Dict[str,Tuple[BiColor,List[str]]] = {}
        extra_node_props : Dict[str,Dict[str,Any]] = {}
        for idx in range(self.my_n):
            my_init_data[f"ext_left{idx+1}"] = (BiColor.RED,[f"wire{idx+1},0"])
            extra_node_props[f"ext_left{idx+1}"] = {"position" : (0,idx+1),
                                                    "my_perfect_edge":-1}
        for idx in range(self.my_n):
            my_init_data[f"ext_right{idx+1}"] =\
                (BiColor.RED,[f"wire{idx+1},{very_last_numbers[idx]-1}"])
            extra_node_props[f"ext_right{idx+1}"] = {"position" : (len(self.my_word)+1,idx+1),
                                                     "my_perfect_edge":-1}
        left_ext = [f"ext_left{idx+1}" for idx in range(self.my_n)]
        right_ext = [f"ext_right{idx+1}" for idx in range(self.my_n)]
        right_ext.reverse()
        external_init_orientation = left_ext + right_ext

        for my_idx,letter_orig in enumerate(self.my_word):
            if letter_orig<0:
                letter = - letter_orig
                bottom_color = BiColor.RED
                top_color = BiColor.GREEN
            else:
                letter = letter_orig
                top_color = BiColor.RED
                bottom_color = BiColor.GREEN
            bottom_number = last_on_this_strand[my_idx][letter-1]
            top_number = last_on_this_strand[my_idx][letter]
            bottom_name = f"wire{letter},{bottom_number}"
            top_name = f"wire{letter+1},{top_number}"
            if bottom_number==0:
                bottom_name_before = f"ext_left{letter}"
            else:
                bottom_name_before = f"wire{letter},{bottom_number-1}"
            if bottom_number==very_last_numbers[letter-1]-1:
                bottom_name_after = f"ext_right{letter}"
            else:
                bottom_name_after = f"wire{letter},{bottom_number+1}"
            if top_number==0:
                top_name_before = f"ext_left{letter+1}"
            else:
                top_name_before = f"wire{letter+1},{top_number-1}"
            if top_number==very_last_numbers[letter]-1:
                top_name_after = f"ext_right{letter+1}"
            else:
                top_name_after = f"wire{letter+1},{top_number+1}"
            my_init_data[bottom_name] = \
                (bottom_color,[bottom_name_before,top_name,bottom_name_after])
            extra_node_props[bottom_name] = {"position" : (my_idx+1,letter),
                                             "my_perfect_edge" : 1}
            my_init_data[top_name] = (top_color,[top_name_before,top_name_after,bottom_name])
            extra_node_props[top_name] = {"position" : (my_idx+1,letter+1),
                                          "my_perfect_edge" : 2}
        my_pb = PlabicGraph(my_init_data,
                 external_init_orientation,
                 {},
                 None,
                 extra_node_props)
        if not my_pb.is_bipartite():
            my_pb.remove_prop("my_perfect_edge")
        return my_pb

if __name__ == "__main__":
    double_word = [-2,1,2,-1,-2,1]
    dw = WiringDiagram(3,double_word)
    p = dw.to_plabic()
    p.draw(draw_oriented_if_perfect = False)

    double_word = [1]
    dw = WiringDiagram(2,double_word)
    p = dw.to_plabic()
    p.draw(draw_oriented_if_perfect = False)

    double_word = [1,-1]
    dw = WiringDiagram(2,double_word)
    p = dw.to_plabic()
    p.draw()



"""
Le Diagram
"""

from enum import Enum, auto
from typing import List, Optional, Tuple, Set, Iterable, Union, TypeVar, Dict, cast, Iterator
import itertools
#pylint:disable=import-error
from .plabic_diagram import BiColor, ExtraData, PlabicGraph

#pylint:disable=too-many-locals,too-many-statements,too-many-branches,too-many-return-statements,too-many-instance-attributes

Point = Tuple[float, float]

FillingInt = Iterable[Iterable[int]]
FillingBool = Iterable[Iterable[bool]]
FourBool = Tuple[bool, bool, bool, bool]
FourInt = Tuple[int, int, int, int]

T = TypeVar("T")
def safe_index(some_l: List[T], idx: int, default: T) -> T:
    """
    use default if would get an IndexError
    """
    return some_l[idx] if idx < len(some_l) else default

class SquareType(Enum):
    """
    the 4 possibilities for the arrow configuration
    at a 1
    """
    GAMMASHAPE = auto()
    TSHAPE = auto()
    SIDETSHAPE = auto()
    PLUSSHAPE = auto()
    JAGGEDVERTICAL = auto()
    JAGGEDHORIZONTAL = auto()


    def to_4_bools(self) -> FourBool:
        """
        where are the arrows on this kind of square
        """
        if self == SquareType.GAMMASHAPE:
            return (False,False,True,True)
        if self == SquareType.TSHAPE:
            return (False,True,True,True)
        if self == SquareType.PLUSSHAPE:
            return (True,True,True,True)
        if self == SquareType.SIDETSHAPE:
            return (True,False,True,True)
        if self == SquareType.JAGGEDVERTICAL:
            return (False,True,False,False)
        if self == SquareType.JAGGEDHORIZONTAL:
            return (True,False,False,False)
        raise ValueError(f"Unreachable, enum exhausted {self}")

def from_arrow_presence(bools : FourBool) -> SquareType:
    """
    which square type is it based on which arrows present
    """
    for sq_type in SquareType:
        if sq_type.to_4_bools() == bools:
            return sq_type
    raise ValueError(f"Impossible configuration of arrow presences {bools}")

def square_type(my_pos : Tuple[int,int],
                all_ones_data : Dict[Tuple[int,int],Tuple[FourBool,FourInt,FourBool]]
                ) -> SquareType:
    """
    the square type at my_pos based on the dictionary provided
    """
    surrounding_1s_exist,_,_ = all_ones_data[my_pos]
    arrow_data = (
        surrounding_1s_exist[0], surrounding_1s_exist[1], True, True)
    return from_arrow_presence(arrow_data)

def surrounding_names(my_pos: Tuple[int, int],
               all_ones_data : Dict[Tuple[int,int],Tuple[FourBool,FourInt,FourBool]]
                ) -> List[Tuple[str, BiColor, List[str],Tuple[float,float]]]:
    """
    the name of the vertex/vertices at my_pos, it's color/colors
    the neighbors in clockwise order, and the position(s) of the
    vertex/vertices at my_pos
    basically the coordinates of my_pos moved from indexing into the tableau
    to where they would be drawn in English convention
    """
    (surrounding_1s_exist,surrounding_1s_locs,_) = all_ones_data[my_pos]

    if surrounding_1s_exist[0]:
        above_loc = (surrounding_1s_locs[0],my_pos[1])
        above_square_type = square_type(above_loc,all_ones_data)
        if above_square_type == SquareType.PLUSSHAPE:
            above_name = f"int{above_loc[0]},{above_loc[1]}_bl"
        else:
            above_name = f"int{above_loc[0]},{above_loc[1]}"
    else:
        above_name = ""

    if surrounding_1s_exist[1]:
        left_loc = (my_pos[0],surrounding_1s_locs[1])
        left_square_type = square_type(left_loc,all_ones_data)
        if left_square_type == SquareType.PLUSSHAPE:
            left_name = f"int{left_loc[0]},{left_loc[1]}_tr"
        else:
            left_name = f"int{left_loc[0]},{left_loc[1]}"
    else:
        left_name = ""

    if surrounding_1s_exist[2]:
        below_loc = (surrounding_1s_locs[2],my_pos[1])
        below_square_type = square_type(below_loc,all_ones_data)
        if below_square_type == SquareType.PLUSSHAPE:
            below_name = f"int{below_loc[0]},{below_loc[1]}_tr"
        else:
            below_name = f"int{below_loc[0]},{below_loc[1]}"
    else:
        below_loc = (surrounding_1s_locs[2],my_pos[1])
        below_name = f"bdry{below_loc[0]},{below_loc[1]}_h"

    if surrounding_1s_exist[3]:
        right_loc = (my_pos[0],surrounding_1s_locs[3])
        right_square_type = square_type(right_loc,all_ones_data)
        if right_square_type == SquareType.PLUSSHAPE:
            right_name = f"int{right_loc[0]},{right_loc[1]}_bl"
        else:
            right_name = f"int{right_loc[0]},{right_loc[1]}"
    else:
        right_loc = (my_pos[0],surrounding_1s_locs[3])
        right_name = f"bdry{right_loc[0]},{right_loc[1]}_v"

    my_square = square_type(my_pos, all_ones_data)
    my_name = f"int{my_pos[0]},{my_pos[1]}"

    if my_square == SquareType.GAMMASHAPE:
        return [(my_name,BiColor.RED,[right_name,below_name],(0.0,0.0))]
    if my_square == SquareType.TSHAPE:
        return [(my_name,BiColor.GREEN,[right_name,below_name,left_name],(0.0,0.0))]
    if my_square == SquareType.PLUSSHAPE:
        my_name_red = f"{my_name}_tr"
        my_name_green = f"{my_name}_bl"
        my_red = (my_name_red,BiColor.RED,[above_name,right_name,my_name_green],(0.1,0.1))
        my_green = (my_name_green,BiColor.GREEN,[my_name_red,below_name,left_name],(-0.1,-0.1))
        return [my_red,my_green]
    if my_square == SquareType.SIDETSHAPE:
        return [(my_name,BiColor.RED,[right_name,below_name,above_name],(0.0,0.0))]
    raise ValueError("Should be unreachable, cases for 1 existing enum exhausted")

class LeDiagram:
    """
    A Young diagram
    with 0,1 filling
    satisfing the
    backwards L condition
    for the position of 1's
    when drawn in English convention
    """
    def __init__(self,pre_filling : Union[FillingInt,FillingBool]):
        """
        give the filling along each row
        left to right and top to bottom
        """
        filling : List[List[bool]] = \
            list(map(lambda cur_row : list(map(lambda z:z>0,cur_row)),pre_filling))
        my_part_lengths = [len(z) for z in filling]
        if sum(my_part_lengths)==0:
            raise ValueError("Expect the shape to be a partition of a positive number")
        for (row_len_a,row_len_b) in zip(my_part_lengths,my_part_lengths[1:]):
            if row_len_a<row_len_b:
                raise ValueError(f"{my_part_lengths} did not describe a Young Diagram")
        self.width = my_part_lengths[0]
        self.height = len(my_part_lengths)
        self.my_diagram = my_part_lengths
        locations_1s : Set[Tuple[int,int]] = set()
        for row_idx, cur_row in enumerate(filling):
            for col_idx, cur_filling in enumerate(cur_row):
                if cur_filling:
                    locations_1s.add((row_idx,col_idx))
        for ((i_1,j_1),(i_2,j_2)) in itertools.combinations(locations_1s,2):
            if i_1<i_2 and j_1>j_2:
                check_here = safe_index(filling[i_2],j_1,True)
                if not check_here:
                    raise ValueError(f"{(i_1,j_1)} and {(i_2,j_2)} were 1, but {(i_2,j_1)} was not")

        self.filling = filling

        ones_data : Dict[Tuple[int,int],Tuple[FourBool,FourInt,FourBool]] = {}
        self.column_heights : List[Optional[int]] = [None]*self.width
        self.nearest_strictly_nw : Dict[Tuple[int,int],Tuple[int,int]] = {}
        self.nearest_weakly_nw : Dict[Tuple[int,int],Tuple[int,int]] = {}

        for (cur_row_idx,cur_col_idx) in locations_1s:

            all_above = \
                list(filling[cur_row_idx-prev_row][cur_col_idx]
                     for prev_row in range(1,cur_row_idx+1))
            has_1_above = any(all_above)
            if has_1_above:
                loc_1_above = all_above.index(True)
                loc_1_above = cur_row_idx - loc_1_above - 1
                above_off_d = False
            else:
                loc_1_above = -1
                above_off_d = True

            all_below = list(safe_index(filling[next_row],cur_col_idx,False)
                                 for next_row in range(cur_row_idx+1,len(filling)))
            has_1_below = any(all_below)
            if has_1_below:
                loc_1_below = all_below.index(True)
                loc_1_below = cur_row_idx + loc_1_below + 1
                below_off_d = False
            else:
                below_off_d = True
                loc_1_below = cast(int,self.column_height(cur_col_idx))

            all_right = list(filling[cur_row_idx][next_col]
                              for next_col in range(cur_col_idx+1,len(filling[cur_row_idx])))
            has_1_right = any(all_right)
            if has_1_right:
                loc_1_right = all_right.index(True)
                loc_1_right = loc_1_right + cur_col_idx+1
                right_off_d = False
            else:
                loc_1_right = len(filling[cur_row_idx])
                right_off_d = True

            all_left = \
                list(filling[cur_row_idx][cur_col_idx-prev_col]
                     for prev_col in range(1,cur_col_idx+1))
            has_1_left = any(all_left)
            if has_1_left:
                loc_1_left = all_left.index(True)
                loc_1_left = cur_col_idx - loc_1_left - 1
                left_off_d = False
            else:
                left_off_d = True
                loc_1_left = -1

            ones_data[(cur_row_idx,cur_col_idx)] =\
                ((has_1_above,has_1_left,has_1_below,has_1_right),
                 (loc_1_above,loc_1_left,loc_1_below,loc_1_right),
                 (above_off_d,left_off_d,below_off_d,right_off_d))
        self.ones_data = ones_data

        cur_box = (0,self.width-1)
        southeast_bdry: List[Tuple[int,int]] = []
        empty_bool_list : List[bool] = []
        for _ in range(self.height+1):
            if cur_box[1]<0 or cur_box[0]>=self.height:
                break
            southeast_bdry.append(cur_box)
            next_len = len(safe_index(self.filling,cur_box[0]+1,empty_bool_list))
            while cur_box[1]+1 > next_len:
                cur_box = (cur_box[0],cur_box[1]-1)
                if cur_box[1]>=0:
                    southeast_bdry.append(cur_box)
            cur_box = (cur_box[0]+1,next_len-1)
        self.southeast_bdry = southeast_bdry

    def column_height(self,which_col : int) -> Optional[int]:
        """
        how many boxes in this column, None which_col is <0 or >=self.width
        """
        if which_col<0 or which_col>=self.width:
            return None
        cached_ht = self.column_heights[which_col]
        if cached_ht is not None:
            return cached_ht
        my_ht = 0
        for my_ht in range(self.height+1):
            try:
                self.filling[my_ht][which_col]
            except IndexError:
                self.column_heights[which_col] = my_ht
                return my_ht
        return None

    def nw_path(self,cur_loc : Tuple[int,int],start_weak : bool) -> Iterator[Tuple[int,int]]:
        """
        path always going strictly northwest starting
        either starts at the nearest 1 strictly/weakly northwest of cur_loc
        """
        my_loc = self.nw_path_next(cur_loc,not start_weak)
        while True:
            if my_loc is None:
                break
            yield my_loc
            my_loc = self.nw_path_next(my_loc,True)

    def nw_path_next(self,cur_loc : Tuple[int,int], strict : bool) -> Optional[Tuple[int,int]]:
        """
        the unique (if exists) box with a 1 in it that is (strictly) to the northwest
        of cur_loc
        """
        manhattan_dist = True
        if manhattan_dist:
            dist_power = 1
        else:
            dist_power = 2
        if strict:
            cached_answer = self.nearest_strictly_nw.get(cur_loc,None)
        else:
            cached_answer = self.nearest_weakly_nw.get(cur_loc,None)
        if cached_answer is not None:
            return cached_answer if cached_answer != (-1,-1) else None
        try:
            my_entry = self.filling[cur_loc[0]][cur_loc[1]]
        except IndexError:
            my_entry = None
        if my_entry is None:
            raise ValueError(f"The current location {cur_loc} must be in the diagram")
        if not strict and my_entry:
            self.nearest_weakly_nw[cur_loc] = cur_loc
            return cur_loc
        if strict and (cur_loc[0]==0 or cur_loc[1]==0):
            self.nearest_strictly_nw[cur_loc] = (-1,-1)
            return None
        # weakly northwest
        if not strict:
            if cur_loc[0]>0:
                option_from_north = self.nw_path_next((cur_loc[0]-1,cur_loc[1]),False)
            else:
                option_from_north = None
            if cur_loc[1]>0:
                option_from_west = self.nw_path_next((cur_loc[0],cur_loc[1]-1),False)
            else:
                option_from_west = None
            if option_from_north is None and option_from_west is None:
                self.nearest_weakly_nw[cur_loc] = (-1,-1)
                return None
            if option_from_north is None:
                self.nearest_weakly_nw[cur_loc] = cast(Tuple[int,int],option_from_west)
                return option_from_west
            if option_from_west is None:
                self.nearest_weakly_nw[cur_loc] = cast(Tuple[int,int],option_from_north)
                return option_from_north
            if option_from_north==option_from_west:
                self.nearest_weakly_nw[cur_loc] = cast(Tuple[int,int],option_from_west)
                return option_from_west
            dist_to_option_from_north = (cur_loc[0]-option_from_north[0])**dist_power+\
                (cur_loc[1]-option_from_north[1])**dist_power
            dist_to_option_from_west = (cur_loc[0]-option_from_west[0])**dist_power+\
                (cur_loc[1]-option_from_west[1])**dist_power
            if dist_to_option_from_north<dist_to_option_from_west:
                self.nearest_weakly_nw[cur_loc] = option_from_north
                return option_from_north
            if dist_to_option_from_north>dist_to_option_from_west:
                self.nearest_weakly_nw[cur_loc] = option_from_west
                return option_from_west
            raise ValueError("There should have been a unique nearest")
        # strictly northwest
        one_step_nw = self.filling[cur_loc[0]-1][cur_loc[1]-1]
        if one_step_nw:
            self.nearest_strictly_nw[cur_loc] = (cur_loc[0]-1,cur_loc[1]-1)
            return (cur_loc[0]-1,cur_loc[1]-1)
        option_from_north = self.nw_path_next((cur_loc[0]-1,cur_loc[1]),True)
        option_from_west = self.nw_path_next((cur_loc[0],cur_loc[1]-1),True)
        if option_from_north is None and option_from_west is None:
            self.nearest_strictly_nw[cur_loc] = (-1,-1)
            return None
        if option_from_north is None:
            self.nearest_strictly_nw[cur_loc] = cast(Tuple[int,int],option_from_west)
            return option_from_west
        if option_from_west is None:
            self.nearest_strictly_nw[cur_loc] = option_from_north
            return option_from_north
        if option_from_north==option_from_west:
            self.nearest_strictly_nw[cur_loc] = option_from_west
            return option_from_west
        dist_to_option_from_north = (cur_loc[0]-option_from_north[0])**dist_power+\
            (cur_loc[1]-option_from_north[1])**dist_power
        dist_to_option_from_west = (cur_loc[0]-option_from_west[0])**dist_power+\
            (cur_loc[1]-option_from_west[1])**dist_power
        if dist_to_option_from_north<dist_to_option_from_west:
            self.nearest_strictly_nw[cur_loc] = option_from_north
            return option_from_north
        if dist_to_option_from_north>dist_to_option_from_west:
            self.nearest_strictly_nw[cur_loc] = option_from_west
            return option_from_west
        raise ValueError("There should have been a unique nearest")

    def _make_jagged_bdry_labels(self,
                                bounding_k : int,
                                bounding_n : int)\
                                      -> List[Tuple[Tuple[int,int],bool]]:
        """
        the lines on the
        jagged northeast->southwest lattice path
        the lines are labelled by the box not on the shape
        they are also adjacent to
        and whether using the left side of this box
        or the top side of this box
        """
        if bounding_k<=0 or bounding_n-bounding_k<=0:
            raise ValueError("Dimensions of bounding box should be positive")
        if bounding_k<self.height or bounding_n-bounding_k<self.width:
            raise ValueError(f"Want to fit inside a {bounding_k} by {bounding_n-bounding_k} box")
        extra_verticals_at_bottom = bounding_k-self.height
        extra_horizontals_at_top = bounding_n-bounding_k-self.width
        cur_box = (0,self.width)
        cur_bdry_vertical = True
        jagged_bdry_labels: List[Tuple[Tuple[int,int],bool]] = []
        empty_bool_list : List[bool] = []
        for _ in range(self.width+self.height):
            jagged_bdry_labels.append((cur_box,cur_bdry_vertical))
            if cur_bdry_vertical:
                if len(safe_index(self.filling,cur_box[0]+1,empty_bool_list))==cur_box[1]:
                    cur_box = (cur_box[0]+1,cur_box[1])
                    cur_bdry_vertical = True
                else:
                    cur_box = (cur_box[0]+1,cur_box[1]-1)
                    cur_bdry_vertical = False
            else:
                if self.column_height(cur_box[1]-1) == cur_box[0]:
                    cur_box = (cur_box[0],cur_box[1]-1)
                    cur_bdry_vertical = False
                else:
                    cur_box = (cur_box[0],cur_box[1])
                    cur_bdry_vertical = True
        jagged_bdry_labels_prepend = \
            [((0,self.width+i),False) for i in range(extra_horizontals_at_top)]
        jagged_bdry_labels_prepend.reverse()
        jagged_bdry_labels_append = \
            [((self.height+i,0),True) for i in range(extra_verticals_at_bottom)]
        return jagged_bdry_labels_prepend+jagged_bdry_labels+jagged_bdry_labels_append

    def _row_col_labels(self,bounding_k : int, bounding_n : int) -> Tuple[List[int],List[int]]:
        """
        the row labels and column labels using the jagged boundary southwest path
        each row of the diagram ends with a row label
        each column of the diagram ends with a column label
        return both of these
        """
        jagged_bdry_labels = self._make_jagged_bdry_labels(bounding_k,bounding_n)
        row_labels = [0]*self.height
        col_labels = [0]*self.width
        idx_in_row_labels = 0
        idx_in_col_labels = 0
        for (box_number,(jagged_bdry_box,left_side_of_it)) in enumerate(jagged_bdry_labels):
            if left_side_of_it and idx_in_row_labels<self.height:
                row_labels[idx_in_row_labels] = box_number+1
                idx_in_row_labels += 1
            elif jagged_bdry_box[1]<self.width:
                col_labels[idx_in_col_labels] = box_number+1
                idx_in_col_labels += 1
        col_labels.reverse()
        return (row_labels,col_labels)

    def to_grassmann_necklace(self,bounding_k : int, bounding_n : int) -> List[Set[int]]:
        """
        to Grassmann necklace
        I_i+1 is only different from I_i in possibly removing i from I_i
        and replacing it by something not in I_i
        each being bounding_k element subsets of 1..bounding_n
        """
        (row_labels,col_labels) = self._row_col_labels(bounding_k, bounding_n)
        return_val : List[Set[int]] = [set()]*bounding_n
        return_val[0] = set(row_labels)
        for (idx,starting_box) in enumerate(self.southeast_bdry):
            if idx+1 not in return_val[idx]:
                return_val[idx+1] = set()
                return_val[idx+1].update(return_val[idx])
                continue
            rows_seen_labels = []
            cols_seen_labels = []
            for cur_box in self.nw_path(starting_box,True):
                cur_row_seen = row_labels[cur_box[0]]
                cur_col_seen = col_labels[cur_box[1]]
                rows_seen_labels.append(cur_row_seen)
                cols_seen_labels.append(cur_col_seen)
            set_i_cur = return_val[0].difference(rows_seen_labels)
            set_i_cur.update(cols_seen_labels)
            return_val[idx+1] = set_i_cur
        return return_val

    @staticmethod
    def from_grassmann_necklace(grassmann_necklace : List[Set[int]],
                                bounding_k : Optional[int]=None,
                                bounding_n : Optional[int]=None) -> "LeDiagram":
        """
        construct Le diagram from a Grassmann necklace
        """
        if bounding_n is None:
            bounding_n = len(grassmann_necklace)
        elif len(grassmann_necklace) != bounding_n:
            raise ValueError(
                f"The Grassmann necklace of type {(bounding_k,bounding_n)} has {bounding_n} sets")
        if bounding_n == 0:
            raise ValueError("Dimensions of bounding box should be positive")
        if bounding_k is None:
            bounding_k = len(grassmann_necklace[0])
        if bounding_k<=0 or bounding_n-bounding_k<=0:
            raise ValueError("Dimensions of bounding box should be positive")
        if any(len(cur_set) != bounding_k for cur_set in grassmann_necklace):
            raise ValueError(
                " ".join([f"In a Grassmann necklace of type {(bounding_k,bounding_n)}",
                          "all sets have {bounding_k} elements"]))
        if any(not cur_set.issubset(range(1,bounding_n+1)) for cur_set in grassmann_necklace):
            raise ValueError(
                " ".join([f"In a Grassmann necklace of type {(bounding_k,bounding_n)}",
                          "all sets are subsets of 1..{bounding_n}"]))
        set_i_1 = grassmann_necklace[0]
        list_i_1 = list(set_i_1)
        list_i_1.sort()
        my_filling : List[List[bool]] = []
        first_part = bounding_n-bounding_k-list_i_1[0]+1
        my_filling.append([False]*first_part)
        prev_row_label = list_i_1[0]
        prev_part_len = first_part
        for cur_row_label in list_i_1[1:]:
            next_part_len = prev_part_len - (cur_row_label-prev_row_label-1)
            my_filling.append([False]*next_part_len)
            prev_row_label = cur_row_label
            prev_part_len = next_part_len

        no_filled_le = LeDiagram(my_filling)
        # pylint:disable=protected-access
        (row_labels,col_labels) = no_filled_le._row_col_labels(bounding_k, bounding_n)
        label_to_row = {}
        for idx,cur_row_label in enumerate(row_labels):
            label_to_row[cur_row_label] = idx
        label_to_col = {}
        for idx,cur_col_label in enumerate(col_labels):
            label_to_col[cur_col_label] = idx

        for set_i_cur in grassmann_necklace[1:]:
            set_i_cur_minus_i_1 = set_i_cur.difference(set_i_1)
            set_i_1_minus_set_i_cur = set_i_1.difference(set_i_cur)
            all_as = list(set_i_1_minus_set_i_cur)
            all_as.sort(reverse=True)
            all_bs = list(set_i_cur_minus_i_1)
            all_bs.sort()
            for (cur_row_label,cur_col_label) in zip(all_as,all_bs):
                which_row = label_to_row[cur_row_label]
                which_col = label_to_col[cur_col_label]
                my_filling[which_row][which_col] = True
        return LeDiagram(my_filling)

    def to_plabic(self) -> PlabicGraph:
        """
        a plabic graph with bridges for each letter
        """

        my_init_data : Dict[str, Tuple[BiColor, List[str]]] = {}
        extra_node_props: Dict[str, ExtraData] = {}

        # on isolated boundary vertices, there is a lollipop
        # what color is the lollipop to make
        # ones for the vertical edges of the jagged path be sources
        # ones for the horizontal edges of the jagged path be sinks
        color_to_make_source = BiColor.RED
        color_to_make_sink = BiColor.GREEN

        for one_pos in self.ones_data:
            my_names_and_neighbors = \
                surrounding_names(one_pos,self.ones_data)
            for my_name,my_color,clockwise_neighbors,pos_shift in my_names_and_neighbors:
                my_init_data[my_name] = (my_color,clockwise_neighbors)
                extra_node_props[my_name] = {"position" :
                                             (one_pos[1]+pos_shift[0],-one_pos[0]+pos_shift[1])}

        cur_box = (0,self.width)
        cur_bdry_type = SquareType.JAGGEDVERTICAL
        external_init_orientation : List[str] = []
        empty_bool_list : List[bool] = []
        for _ in range(self.width+self.height):
            this_bdry_non_lollipop = True
            if cur_bdry_type == SquareType.JAGGEDVERTICAL:
                my_name = f"bdry{cur_box[0]},{cur_box[1]}_v"
                cur_list = self.filling[cur_box[0]].copy()
                cur_list.reverse()
                try:
                    pos_last_1 = cur_box[1] - cur_list.index(True) - 1
                except ValueError:
                    this_bdry_non_lollipop = False
                if this_bdry_non_lollipop:
                    last_in_this_row = f"int{cur_box[0]},{pos_last_1}"
                    last_in_this_row_type = square_type((cur_box[0],pos_last_1),self.ones_data)
                    if last_in_this_row_type == SquareType.PLUSSHAPE:
                        last_in_this_row = f"{last_in_this_row}_tr"
                    my_init_data[my_name] = (BiColor.RED,[last_in_this_row])
                    my_position = (cur_box[1]-0.5,-cur_box[0]+0.0)
                    extra_node_props[my_name] = {"position" : my_position}
                else:
                    my_lollipop_name = f"{my_name}_lollipop"
                    my_init_data[my_name] = (BiColor.RED,[my_lollipop_name])
                    my_position = (cur_box[1]-0.5,-cur_box[0]+0.0)
                    extra_node_props[my_name] = {"position" : my_position}
                    my_init_data[my_lollipop_name] = (color_to_make_source,[my_name])
                    my_position = (cur_box[1]-0.7,-cur_box[0]+0.0)
                    extra_node_props[my_lollipop_name] = {"position" : my_position}
                if len(safe_index(self.filling,cur_box[0]+1,empty_bool_list))==cur_box[1]:
                    cur_box = (cur_box[0]+1,cur_box[1])
                    cur_bdry_type = SquareType.JAGGEDVERTICAL
                else:
                    cur_box = (cur_box[0]+1,cur_box[1]-1)
                    cur_bdry_type = SquareType.JAGGEDHORIZONTAL
            else:
                my_name = f"bdry{cur_box[0]},{cur_box[1]}_h"
                pos_last_1 = cur_box[0]-1
                while pos_last_1>=0 and not self.filling[pos_last_1][cur_box[1]]:
                    pos_last_1 -= 1
                if pos_last_1 == -1:
                    this_bdry_non_lollipop = False
                if this_bdry_non_lollipop:
                    last_in_this_col = f"int{pos_last_1},{cur_box[1]}"
                    last_in_this_col_type = square_type((pos_last_1,cur_box[1]),self.ones_data)
                    if last_in_this_col_type == SquareType.PLUSSHAPE:
                        last_in_this_col = f"{last_in_this_col}_bl"
                    my_init_data[my_name] = (BiColor.RED,[last_in_this_col])
                    my_position = (cur_box[1]+0.0,-cur_box[0]+0.5)
                    extra_node_props[my_name] = {"position" : my_position}
                else:
                    my_lollipop_name = f"{my_name}_lollipop"
                    my_init_data[my_name] = (BiColor.RED,[my_lollipop_name])
                    my_position = (cur_box[1]+0.0,-cur_box[0]+0.5)
                    extra_node_props[my_name] = {"position" : my_position}
                    my_init_data[my_lollipop_name] = (color_to_make_sink,[my_name])
                    my_position = (cur_box[1]+0.0,-cur_box[0]+0.7)
                    extra_node_props[my_lollipop_name] = {"position" : my_position}
                if self.column_height(cur_box[1]-1) == cur_box[0]:
                    cur_box = (cur_box[0],cur_box[1]-1)
                    cur_bdry_type = SquareType.JAGGEDHORIZONTAL
                else:
                    cur_box = (cur_box[0],cur_box[1])
                    cur_bdry_type = SquareType.JAGGEDVERTICAL
            external_init_orientation.append(my_name)

        return PlabicGraph(my_init_data,
                 external_init_orientation,
                 {},
                 [],
                 extra_node_props)

    def __str__(self) -> str:
        return "\n".join([
            " ".join(["1" if z else "0" for z in cur_row])
            for cur_row in self.filling])

    def __eq__(self,other) -> bool:
        if not isinstance(other,LeDiagram):
            return False
        return self.filling == other.filling

if __name__ == "__main__":

    # Example from
    # https://doi.org/10.1016/j.jcta.2018.06.001
    my_diagram = [[0,1,0,1,0],[1,1,0,1],[0,0],[0,1]]
    my_Le = LeDiagram(my_diagram)
    print(my_Le)
    print(my_Le.ones_data)
    print(my_Le.column_height(0))
    # pylint:disable=protected-access
    print(my_Le._make_jagged_bdry_labels(4,9))
    print(my_Le._make_jagged_bdry_labels(5,10))
    print(my_Le._make_jagged_bdry_labels(7,15))
    print(f"Path starting at {(0,4)} is {list(my_Le.nw_path((0,4),True))}")
    print(f"Path starting at {(1,3)} is {list(my_Le.nw_path((1,3),True))}")
    print(f"Path starting at {(2,1)} is {list(my_Le.nw_path((2,1),True))}")
    print(f"Path starting at {(3,1)} is {list(my_Le.nw_path((3,1),True))}")
    print(f"Southeast boundary : {my_Le.southeast_bdry}")
    print(f"Grassmann necklace : {my_Le.to_grassmann_necklace(4,9)}")
    p = my_Le.to_plabic()
    p.draw(show_node_names=False)

    # case of n by k rectangle with all 1's
    # have particularly nice face quivers
    k,n = 5,4
    my_diagram = [[1]*k]*n
    my_Le = LeDiagram(my_diagram)
    p = my_Le.to_plabic()
    p.draw(show_node_names=False)



"""
test for double wiring diagrams
"""
#pylint:disable=invalid-name,import-error
from typing import List
from plabic import WiringDiagram

def test_dw1() -> None:
    """
    sbar_2 s_1 s2 sbar_1 sbar_2 s1
    """
    double_word = [-2,1,2,-1,-2,1]
    dw = WiringDiagram(3,double_word)
    expected_chamber_minors = [
        (set([1,2]),set([2,3])),
        (set([1]),set([3])),
        (set([1,2]),set([1,3])),
        (set([2]),set([3])),
        (set([2,3]),set([1,3])),
        (set([2]),set([1])),
        (set([3]),set([1])),
        (set([2,3]),set([1,2])),
        (set([1,2,3]),set([1,2,3]))
    ]
    for expected_chamber_minor,chamber_minor in zip(expected_chamber_minors,dw.chamber_minors()):
        assert chamber_minor == expected_chamber_minor
    p = dw.to_plabic()
    assert p.my_extra_props == set(["position"])
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert did_scale_up

def test_dw2() -> None:
    """
    empty word
    """
    DID_ERROR = False
    try:
        double_word : List[int] = []
        _dw = WiringDiagram(2,double_word)
    except ValueError:
        DID_ERROR = True
    assert DID_ERROR

def test_dw3() -> None:
    """
    s_1
    """
    double_word = [1]
    dw = WiringDiagram(2,double_word)
    p = dw.to_plabic()
    assert p.is_bipartite()
    assert p.my_extra_props == set(["position","my_perfect_edge"])
    assert p.my_perfect_matching == set([("wire1,0","wire2,0",1),("wire2,0","wire1,0",2)])
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert did_scale_up

def test_dw4() -> None:
    """
    s_1 sbar_1
    """
    double_word = [1,-1]
    dw = WiringDiagram(2,double_word)
    p = dw.to_plabic()
    assert p.my_extra_props == set(["position","my_perfect_edge"])
    assert p.my_perfect_matching == set([("wire1,0","wire2,0",1),
                                         ("wire2,0","wire1,0",2),
                                         ("wire1,1","wire2,1",1),
                                         ("wire2,1","wire1,1",2)])
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert did_scale_up



"""
test Le diagrams
"""

#pylint:disable=invalid-name,import-error
from typing import Set
from plabic import LeDiagram

def test_empty() -> None:
    """
    partition should be of positive number
    """
    try:
        _my_Le = LeDiagram([])
    except ValueError as e:
        assert str(e)=="Expect the shape to be a partition of a positive number"
    try:
        _my_Le = LeDiagram([[],[],[]])
    except ValueError as e:
        assert str(e)=="Expect the shape to be a partition of a positive number"

def test_Le1() -> None:
    """
    Example from
    https://doi.org/10.1016/j.jcta.2018.06.001
    """
    my_diagram = [[0,1,0,1,0],[1,1,0,1],[0,0],[0,1]]
    my_Le = LeDiagram(my_diagram)
    assert my_Le.filling == my_diagram
    print(my_Le.column_height(0) == 4)
    print(my_Le.column_height(1) == 4)
    print(my_Le.column_height(2) == 2)
    print(my_Le.column_height(3) == 2)
    print(my_Le.column_height(4) == 1)
    print(my_Le.column_height(5) == 0)
    p = my_Le.to_plabic()
    assert p.my_extra_props == set(["position"])
    assert p.my_perfect_matching is None
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert did_scale_up

def test_rectangle() -> None:
    """
    case of n by k rectangle with all 1's
    have particularly nice face quivers
    """
    k,n = 5,4
    my_diagram = [[1]*k]*n
    my_Le = LeDiagram(my_diagram)
    p = my_Le.to_plabic()
    assert p.my_extra_props == set(["position"])
    assert p.my_perfect_matching is None
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert did_scale_up

def from_multidigit_int(multi_digit : int) -> Set[int]:
    """
    when n is <=9
    can give the subset of 1..n
    as just concatenation
    """
    return {int(i) for i in str(multi_digit)}

def test_to_Grassmann() -> None:
    """
    example to Grassmann necklace
    from https://arxiv.org/pdf/1803.01726.pdf
    """
    my_diagram = [[1,1,0,0,1],[1,1,0,1],[0,1]]
    my_Le = LeDiagram(my_diagram)
    obs_necklace = my_Le.to_grassmann_necklace(3,8)
    exp_necklace_nums = [136,236,367,467,678,678,178,168]
    exp_necklace = [from_multidigit_int(z) for z in exp_necklace_nums]
    exp_necklace_2 = [set([1,3,6]),set([2,3,6]),set([3,6,7]),
                    set([4,6,7]),set([6,7,8]),set([6,7,8]),
                    set([1,7,8]),set([1,6,8])]
    assert exp_necklace == exp_necklace_2
    assert obs_necklace == exp_necklace
    p = my_Le.to_plabic()
    assert p.my_extra_props == set(["position"])
    assert p.my_perfect_matching is None
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert did_scale_up

def test_from_Grassmann() -> None:
    """
    example from Grassmann necklace
    from https://arxiv.org/pdf/1803.01726.pdf
    """
    necklace_nums = [1247, 2347, 3478, 4678, 5678, 4678, 1478, 1478]
    necklace = [from_multidigit_int(z) for z in necklace_nums]
    print(necklace)
    my_Le = LeDiagram.from_grassmann_necklace(necklace,4,8)
    print(my_Le)
    exp_Le = LeDiagram([[1,0,0,1],[1,1,0,1],[0,0,1],[0]])
    assert my_Le.filling == exp_Le.filling
    p = my_Le.to_plabic()
    assert p.my_extra_props == set(["position"])
    assert p.my_perfect_matching is None
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert did_scale_up



"""
a framed little 2-disk configuration
"""
from __future__ import annotations
from typing import Callable,Tuple,List,Optional,cast
import numpy as np
import matplotlib.pyplot as plt

Point = Tuple[float,float]
Radius = float
Angle = float

class FramedDiskConfig:
    """
    a framed little 2-disk configuration
    """

    def __init__(self,
                 internal_circles : List[Tuple[Point,Radius,Angle]],
                 outer_circle : Optional[Tuple[Point,Radius,Angle]] = None):
        if outer_circle is None:
            outer_circle = ((0.0,0.0),1.0,0.0)
        self.outer_circle = outer_circle
        self.internal_circles = internal_circles.copy()

    # pylint:disable=too-many-locals
    def do_transformation(self,which_disk : int,
                           other : FramedDiskConfig) -> Callable[[Point],Point]:
        """
        the transformation necessary to take which_disk internal of self
        to the outer_circle of other
        and does that on self
        """
        if which_disk < 0:
            raise ValueError("Which disk is a natural number")
        if len(self.internal_circles) <= which_disk:
            raise ValueError("Not enough internal disks")
        target_center, target_radius, target_angle = other.outer_circle
        cur_center, cur_radius, cur_angle = self.internal_circles[which_disk]
        as_np_cur_center = np.array(cur_center)
        scale_factor = target_radius/cur_radius
        angle_diff = target_angle-cur_angle
        cos_angle_diff, sin_angle_diff = \
            np.cos(angle_diff), np.sin(angle_diff)
        rot_matrix = np.array(((cos_angle_diff, -sin_angle_diff),
                                (sin_angle_diff, cos_angle_diff)))
        as_np_target_center = np.array(target_center)
        def point_transformation(my_point : Point) -> Point:
            as_np_mine = np.array(my_point)
            centered_pt = (as_np_mine - as_np_cur_center)*scale_factor
            centered_rotated_pt = np.matmul(rot_matrix,centered_pt)
            final_pt = centered_rotated_pt + as_np_target_center
            return cast(Point,tuple(final_pt.tolist()))
        def circle_transformation(framed_circle : Tuple[Point,Radius,Angle])\
            -> Tuple[Point,Radius,Angle]:
            (old_center,old_radius,old_angle) = framed_circle
            new_center = point_transformation(old_center)
            assert isinstance(new_center,tuple) and len(new_center)==2 and\
                isinstance(new_center[0],float) and isinstance(new_center[1],float)
            return (new_center,
                    old_radius*scale_factor,
                    old_angle+angle_diff)
        self.outer_circle = circle_transformation(self.outer_circle)
        _ = self.internal_circles.pop(which_disk)
        self.internal_circles = [circle_transformation(z) for z in self.internal_circles]
        self.internal_circles.extend(other.internal_circles)
        return point_transformation

    #pylint:disable=too-many-arguments
    def draw(self,on_top_of_existing : bool = False,
             show_as_well : bool = True,
             draw_framing : bool = True,
             external_circle_color : str="black",
             internal_circles_color : str="black"):
        """
        draw circles
        """
        if on_top_of_existing:
            fig = plt.gcf()
            my_ax = fig.gca()
        else:
            fig, my_ax = plt.subplots()
        (center_x,center_y),radius,_ = self.outer_circle
        if len(self.internal_circles)==0 or not draw_framing:
            min_radius = radius
        else:
            min_radius = min(internal_radius for _,internal_radius,_ in self.internal_circles)
        dot_radius = min_radius / 10
        plt.ylim(center_y-radius-2*dot_radius,center_y+radius+2*dot_radius)
        plt.xlim(center_x-radius-2*dot_radius,center_x+radius+2*dot_radius)

        def draw_a_circle(framed_circle : Tuple[Point,Radius,Angle],color : str):
            (center_x,center_y),radius,angle = framed_circle
            circle = plt.Circle((center_x, center_y), radius, color=color,fill=False)
            my_ax.add_patch(circle)
            if draw_framing:
                cos_angle, sin_angle = \
                    np.cos(angle), np.sin(angle)
                rot_matrix = np.array(((cos_angle, -sin_angle),
                                        (sin_angle, cos_angle)))
                marked_point = np.array([center_x,center_y])+\
                    radius*np.matmul(rot_matrix,np.array((1.0,0.0)))
                marked_point_list = marked_point.tolist()
                xmarked,ymarked = marked_point_list[0], marked_point_list[1]
                circle2 = plt.Circle((xmarked, ymarked), dot_radius, color=color,fill=True)
                my_ax.add_patch(circle2)

        draw_a_circle(self.outer_circle,external_circle_color)
        for cur_internal in self.internal_circles:
            draw_a_circle(cur_internal,internal_circles_color)

        plt.draw()
        if not on_top_of_existing and show_as_well:
            plt.show()

if __name__ == '__main__':
    e = FramedDiskConfig([])
    e.draw()

    e1 = FramedDiskConfig([((0,.5),.2,0)])
    e1.draw()

    e2 = FramedDiskConfig([((0,.5),.2,0),((-.5,0),.3,3.14)])
    e2.draw()

    _ = e1.do_transformation(0,e2)
    e1.draw(draw_framing=False)

    _ = e1.do_transformation(0,e2)
    e1.draw(draw_framing=False)



"""
diagrams used in the sense of cluster algebra
"""

from enum import Enum, auto
from typing import Iterator, List, Set, Tuple, TypeVar, Union, Callable, Optional
from typing import runtime_checkable, Protocol, Any
import itertools
from functools import reduce
import networkx as nx

T = TypeVar("T")

class Arithmetic(Protocol):
    """
    can be multiplied and added
    """
    def __mul__(self : T,other : T) -> T: ...
    def __add__(self : T,other : T) -> T: ...

NT = TypeVar("NT",bound=Arithmetic)

WeightType = Union[float, NT]
EdgeData = Tuple[int, int, WeightType]
Nat = int

class ChipType(Enum):
    """
    which direction the line on a chip goes
    """
    UP = auto()
    DOWN = auto()
    FLAT = auto()


ChipWord = List[Tuple[ChipType, Nat]]

# pylint:disable=line-too-long,invalid-name,too-many-locals,no-member,too-many-branches,too-few-public-methods


class PlanarNetwork:
    """
    built on top of a directed graph
    """

    @staticmethod
    def from_chip_word(chip_word: ChipWord, weights: List[WeightType]) -> List[List[EdgeData]]:
        """
        edge data as in initializer
        the initializer doesn't take this because we also want to
        present the diagram as it would be drawn
        with multiple elementary pieces occuring in the same stage
        provided they concern different indices
        """
        if len(weights) != len(chip_word):
            raise ValueError("The number of weights and chips must match")
        ret_val = []
        for ((ct, idx), wt) in zip(chip_word, weights):
            if ct == ChipType.UP:
                cur_edge = (idx, idx+1, wt)
            elif ct == ChipType.DOWN:
                cur_edge = (idx-1, idx, wt)
            else:
                cur_edge = (idx, idx, wt)
            ret_val.append([cur_edge])
        return ret_val

    def __init__(self, n: Nat, *, edge_list: Optional[List[List[EdgeData]]] = None,
                 chip_word: Optional[ChipWord] = None, chip_weights: Optional[List[WeightType]] = None,
                 multiplicative_identity: WeightType = 1.0,
                 additive_identity: WeightType = 0.0,
                 totally_connected: bool = True):
        """
        a planar diagram with n horizontal lines
        """
        self._underlying_graph = nx.DiGraph()
        self.multiplicative_identity = multiplicative_identity
        self.additive_identity = additive_identity
        self.totally_connected = totally_connected
        for x in range(1, n+1):
            self._underlying_graph.add_node(f"{x}_0")
        sink_number = {x: 0 for x in range(1, n+1)}
        if edge_list is None:
            if chip_word is not None and chip_weights is not None:
                edge_list = PlanarNetwork.from_chip_word(
                    chip_word, chip_weights)
            else:
                raise ValueError(
                    "Either edge_list must be given or both chip_word and chip_weights")
        self.chip_type: Optional[ChipWord] = [] if chip_word is None else chip_word
        for stage in edge_list:
            already_incremented = []
            for edge in stage:
                from_node, to_node, cur_weight = edge
                if chip_word is None and self.chip_type is not None:
                    if from_node-to_node == 1:
                        self.chip_type.append((ChipType.DOWN, to_node))
                    elif to_node - from_node == 1:
                        self.chip_type.append((ChipType.UP, to_node))
                    elif to_node == from_node:
                        self.chip_type.append((ChipType.FLAT, to_node))
                    else:
                        self.chip_type = None
                if from_node in already_incremented:
                    from_number = sink_number[from_node]-1
                else:
                    from_number = sink_number[from_node]
                if to_node in already_incremented:
                    to_number = sink_number[to_node]
                else:
                    to_number = sink_number[to_node]+1
                    sink_number[to_node] += 1
                    self._underlying_graph.add_edge(f"{to_node}_{to_number-1}",
                                                    f"{to_node}_{to_number}", weight=multiplicative_identity)
                    already_incremented.append(to_node)
                self._underlying_graph.add_node(f"{to_node}_{to_number}")
                self._underlying_graph.add_edge(f"{from_node}_{from_number}",
                                                f"{to_node}_{to_number}", weight=cur_weight)
        self.sink_number = sink_number

    def path_weight(self, path_nodes: List[Any]) -> WeightType:
        """
        multiply the weights for the path that connects all the path_nodes
        """
        path_edge_weights : Iterator[WeightType] = (self._underlying_graph[path_nodes[i]][path_nodes[i+1]]['weight']
                      for i in range(len(path_nodes)-1))
        return reduce(lambda acc, y: acc*y, (z for z in path_edge_weights), self.multiplicative_identity)

    def weight_matrix(self, my_i: Nat, my_j: Nat) -> WeightType:
        """
        the sum of the path weights over all simple paths connecting
        the source node on horizontal line i
        the sink node on horizontal line j
        """
        source_node = f"{my_i}_0"
        sink_node = f"{my_j}_{self.sink_number[my_j]}"
        weight_ij = self.additive_identity
        for path_nodes in nx.all_simple_paths(self._underlying_graph, source_node, sink_node):
            cur_contrib = self.path_weight(path_nodes)
            weight_ij += cur_contrib
        return weight_ij

    def vertex_disjoint_collection(self, i_set: Set[Nat], j_set: Set[Nat]) -> Iterator[WeightType]:
        """
        iterate through all systems of paths
        that connect the sources with i_set to the sinks with j_set
        suppose there are k of each so we get k paths each time
        skip if they are not vertex disjoint
        but if they are vertex disjoint give the product of the path weights
        """
        # start off by finding the possibilities of how i_set and j_set match up
        # so that (i,j) in collection means that particular i in i_set connects to j in j_set
        if len(i_set) == 0 or len(j_set) == 0:
            raise ValueError("The sets of sources and sinks must be nonempty")
        relevant_pairings = [pairings_nc(i_set, j_set)]
        # relevant_pairings = pairings(i_set,j_set)
        for collection in relevant_pairings:
            path_collection_iterator = None
            # build up path_collection_iterator by taking the product of iterators
            # where each factor iterator is the ones that connect i to j
            ij_paths = (nx.all_simple_paths(
                self._underlying_graph, f"{cur_i}_0", f"{cur_j}_{self.sink_number[cur_j]}")
                for cur_i, cur_j in collection)
            path_collection_iterator = itertools.product(*ij_paths)
            # now path_collection is a tuple of k paths that fully connect i_set to j_set
            for path_collection in path_collection_iterator:
                seen_vertices: Set[Nat] = set({})
                is_vertex_disjoint = True
                my_paths = []
                for path in path_collection:
                    my_paths.append(list(path))
                    seen_vertices.isdisjoint(set(path))
                    if seen_vertices.isdisjoint(set(path)):
                        seen_vertices = seen_vertices.union(set(path))
                    else:
                        is_vertex_disjoint = False
                        break
                if is_vertex_disjoint:
                    all_path_weights = [self.path_weight(
                        path_nodes) for path_nodes in my_paths]
                    yield reduce(lambda x, y: x*y, all_path_weights, self.multiplicative_identity)

    def lindstrom_minor(self, i_set: Set[Nat], j_set: Set[Nat]) -> WeightType:
        """
        the minor where specify to include the rows and columns specified by i_set and j_set
        """
        minor_value = self.additive_identity
        had_contribution = False
        for contrib in self.vertex_disjoint_collection(i_set, j_set):
            had_contribution = True
            minor_value += contrib
        if not had_contribution:
            self.totally_connected = False
        return minor_value

    def totally_nonnegative(self, nonnegative_wt: Callable[[NT], bool]) -> bool:
        """
        is the associated weight matrix manifestly totally nonnegative
        by the constituent weights being all nonnegative
        """
        all_weights : Iterator[WeightType] = (self._underlying_graph[src][tgt]['weight']
                       for src, tgt in self._underlying_graph.edges())
        def real_nonnegative_wt(arg : Union[float,NT]) -> bool:
            """
            essentially same as nonnegative_wt
            """
            if isinstance(arg,(float,int)):
                return arg>=0
            return nonnegative_wt(arg)
        return all((real_nonnegative_wt(wt) for wt in all_weights))

    def positive(self, positive_wt: Callable[[NT], bool]) -> bool:
        """
        is the associated weight matrix manifestly totally positive
        by the constituent weights being all positive
        """
        assert self.totally_connected
        all_weights : Iterator[WeightType] = (self._underlying_graph[src][tgt]['weight']
                       for src, tgt in self._underlying_graph.edges())
        def real_positive_wt(arg : Union[float,NT]) -> bool:
            """
            essentially same as positive_wt
            """
            if isinstance(arg,(float,int)):
                return arg>=0
            return positive_wt(arg)
        return all((real_positive_wt(wt) for wt in all_weights))

U = TypeVar("U")

def pairings(i_set: Set[T], j_set: Set[U]) -> Iterator[Set[Tuple[T, U]]]:
    """
    generator that goes through all pairings of i_set with j_set
    """
    if len(i_set) != len(j_set):
        raise ValueError("Must have same number of elements")
    if len(i_set) == 0:
        yield set({})
        return
    xtemp = i_set.copy()
    i_popped = xtemp.pop()
    for my_i, my_j in {(i_popped, j) for j in j_set}:
        ytemp = j_set.copy()
        ytemp.remove(my_j)
        for remaining in pairings(xtemp, ytemp):
            remaining.add((my_i, my_j))
            yield remaining


@runtime_checkable
class Ordered(Protocol):
    """
    protocol for ordered
    """

    def __leq__(self, _other): ...


V = TypeVar('V', bound=Ordered)
W = TypeVar('W', bound=Ordered)


def pairings_nc(i_set: Set[V], j_set: Set[W]) -> Set[Tuple[V, W]]:
    """
    the unique pairing of i_set with j_set
    such that the minimum of i_set is matched with the minimum of j_set
    and so on all the way up
    """
    if len(i_set) != len(j_set):
        raise ValueError("Must have same number of elements")
    i_list = list(i_set)
    i_list.sort()
    j_list = list(j_set)
    j_list.sort()
    return set(zip(i_list, j_list))

# pylint:disable = too-many-arguments
def determinant(matrix : List[List[NT]], mul : NT,
                simplifier : Callable[[NT],NT],
                additive_identity : NT,
                negative_one : NT,
                simplify_freq : int=3,
                original_width : Optional[int] =None) -> NT:
    """
    determinant using only + and *
    so that it can work on complex object types
    that override __add__ and __mul__
    imported det functions assume primitive numeric type like float
    """
    width = len(matrix)
    if original_width is None:
        original_width = width
    if width == 1:
        to_return = mul * matrix[0][0]
        if original_width == width:
            to_return = simplifier(to_return)
        return to_return
    sign = negative_one
    answer = additive_identity
    for skip_col in range(width):
        m = []
        for jdx in range(1, width):
            buff = []
            for kdx in range(width):
                if kdx != skip_col:
                    buff.append(matrix[jdx][kdx])
            m.append(buff)
        sign *= negative_one
        answer = answer + mul * \
            determinant(m, sign * matrix[0][skip_col], simplifier,
                        additive_identity, negative_one, simplify_freq, original_width)
    if width % simplify_freq == original_width % simplify_freq:
        answer = simplifier(answer)
    return answer



"""
test triangulations of convex m-gons
"""
#pylint:disable=import-error,invalid-name,R0801
from math import pi as PI,sin,cos
from plabic import Triangulation

def test_octagon() -> None:
    """
    a triangulation of an octagon
    """
    N = 8
    my_diagonals = [(4,7),(2,8),(2,4),(2,7),(4,6)]
    RADIUS = 2
    t = Triangulation([ (RADIUS*cos(2*PI*which/N),RADIUS*sin(2*PI*which/N))
                       for which in range(N)], [(x-1,y-1) for x,y in my_diagonals])
    diags_before = t.my_diagonal_list.copy()
    for diag,quad in zip(t.my_diagonal_list,t.my_quadrilaterals):
        assert diag[0]==quad[0]
        assert diag[1]==quad[2]
    p = t.to_plabic()
    assert p.my_extra_props == set(["position","my_perfect_edge"])
    assert p.my_perfect_matching is not None
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*1.10,z[1]*1.1))
    assert did_scale_up
    change, new_diag = t.quad_flip((1,3))
    assert change
    assert new_diag == (2,6)
    for which_diag_now,diag in enumerate(t.my_diagonal_list):
        if diag != (2,6):
            assert diag == diags_before[which_diag_now]
        else:
            assert diags_before[which_diag_now] == (1,3)
    for diag,quad in zip(t.my_diagonal_list,t.my_quadrilaterals):
        assert diag[0]==quad[0]
        assert diag[1]==quad[2]
    p = t.to_plabic()
    assert p.my_extra_props == set(["position","my_perfect_edge"])
    assert p.my_perfect_matching is not None
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*1.3,z[1]*2.4))
    assert did_scale_up

def test_pentagon() -> None:
    """
    a triangulation of an pentagon
    """
    N = 5
    my_diagonals = [(1,3),(1,4)]
    RADIUS = 2
    t = Triangulation([ (RADIUS*cos(2*PI*which/N),RADIUS*sin(2*PI*which/N))
                       for which in range(N)], [(x-1,y-1) for x,y in my_diagonals])
    diags_before = t.my_diagonal_list.copy()
    for diag,quad in zip(t.my_diagonal_list,t.my_quadrilaterals):
        assert diag[0]==quad[0]
        assert diag[1]==quad[2]
    p = t.to_plabic()
    assert p.my_extra_props == set(["position","my_perfect_edge"])
    assert p.my_perfect_matching is not None
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*1.50,z[1]*1.1))
    assert did_scale_up
    change, new_diag = t.quad_flip((0,2))
    assert change
    assert new_diag == (1,3)
    for which_diag_now,diag in enumerate(t.my_diagonal_list):
        if diag != (1,3):
            assert diag == diags_before[which_diag_now]
        else:
            assert diags_before[which_diag_now] == (0,2)
    for diag,quad in zip(t.my_diagonal_list,t.my_quadrilaterals):
        assert diag[0]==quad[0]
        assert diag[1]==quad[2]
    p = t.to_plabic()
    assert p.my_extra_props == set(["position","my_perfect_edge"])
    assert p.my_perfect_matching is not None
    did_scale_up = p.coordinate_transform(lambda z: (z[0]*1.3,z[1]*2.4))
    assert did_scale_up



"""
plabic graph
PLAnar BIColored
"""
# pylint:disable=too-many-lines
from __future__ import annotations
from enum import Enum, auto
import numbers
from typing import Tuple, Optional, List, Dict, cast, Any, Callable, Set
import itertools
import networkx as nx
import matplotlib.pyplot as plt
# pylint:disable=import-error
from .framed_2_disks import FramedDiskConfig

Point = Tuple[float,float]

class BiColor(Enum):
    """
    the two colors for internal vertices
    """
    RED = auto()
    GREEN = auto()


ExtraData = Dict[str, Any]

# pylint:disable = too-many-instance-attributes
class PlabicGraphBuilder:
    """
    PlabicGraph constructor arguments are complicated to offload this process
    to PlabicGraphBuilder
    """
    my_init_data : Dict[str, Tuple[BiColor, List[str]]]
    num_external_vertices : int
    is_set_externals : List[bool]
    external_init_orientation : List[str]
    multi_edge_permutation : Dict[Tuple[str, str], Dict[int, int]]
    internal_circles_nums : List[int]
    is_set_internals : List[List[bool]]
    internal_bdry_orientations: Optional[List[List[str]]]
    extra_node_props: Optional[Dict[str, ExtraData]]
    have_multi_edges : Set[Tuple[str,str]]

    def __init__(self):
        self.my_init_data = {}
        self.num_external_vertices = 0
        self.is_set_externals = []
        self.external_init_orientation = []
        self.multi_edge_permutation = {}
        self.internal_circles_nums = []
        self.is_set_internals = []
        self.internal_bdry_orientations = None
        self.extra_node_props = None
        self.have_multi_edges = set()

    def set_num_external(self,num_ext : int) -> None:
        """
        set the number of vertices on the outside boundary
        """
        if num_ext<0:
            raise ValueError("Num external vertices should be nonnegative")
        self.num_external_vertices = num_ext
        self.external_init_orientation = [""]*num_ext
        self.is_set_externals = [False]*num_ext

    def set_internal_circles_nums(self,internal_circles_nums : List[int]):
        """
        set the number of internal boundaries and how many vertices are on each
        """
        if any(z<0 for z in internal_circles_nums):
            raise ValueError("Num vertices on an internal boundary should all be nonnegative")
        self.internal_circles_nums = internal_circles_nums
        self.is_set_internals = [[False]*z for z in internal_circles_nums]
        self.internal_bdry_orientations = [[""]*z for z in internal_circles_nums]

    def set_extras(self, vertex_name : str, extras : Optional[ExtraData] = None):
        """
        set the extra_node_props for a specific vertex
        """
        if extras is not None:
            if self.extra_node_props is None:
                self.extra_node_props = {}
            self.extra_node_props[vertex_name] = extras

    def add_internal_vertex(self,vertex_name : str, color : BiColor,
                            clockwise_nhbrs : List[str],
                            extras : Optional[ExtraData] = None) -> None:
        """
        add an internal vertex by it's color and it's half-edges around it
        the half-edges are in clockwise order, they give the vertex at the other
        end
        any extra information want on that node is also provided
        """
        for name in clockwise_nhbrs:
            if clockwise_nhbrs.count(name)>1:
                self.have_multi_edges.add((vertex_name,name))
                self.have_multi_edges.add((name,vertex_name))
        self.my_init_data[vertex_name] = (color,clockwise_nhbrs)
        self.set_extras(vertex_name,extras)

    def add_external_bdry_vertex(self,external_vertex_name : str,
                            idx : int, my_internal_connection : str,
                            extras : Optional[ExtraData] = None) -> None:
        """
        add a boundary vertex on the external boundary by it's only neighbor
        any extra information want on that node is also provided
        it is placed on the external boundary in position provided by idx
        you must have done set_num_external before this so we know what
        range idx is regarded as taking values in
        """
        if idx<0 or idx>=self.num_external_vertices:
            raise ValueError(f"{self.num_external_vertices} external vertices allowed")
        self.my_init_data[external_vertex_name] = (BiColor.RED,[my_internal_connection])
        self.external_init_orientation[idx] = external_vertex_name
        self.is_set_externals[idx] = True
        self.set_extras(external_vertex_name,extras)

    # pylint:disable=too-many-arguments
    def add_internal_bdry_vertex(self,internal_bdry_vertex_name : str,
                            which_circle : int,
                            where_in_that_circle : int,
                            my_internal_connection : str,
                            extras : Optional[ExtraData] = None) -> None:
        """
        add a boundary vertex on an internal boundary by it's only neighbor
        any extra information want on that node is also provided
        it is placed on the internal boundary which_circle in position provided by
        where_in_that_circle
        you must have done set_internal_circles_nums before this so we know what
        range which_circle and where_in_that_circle are regarded as taking values in
        """
        if self.internal_bdry_orientations is None:
            raise ValueError(
                "Should have set the number and sizes of the internal boundaries already")
        if which_circle<0 or which_circle>=len(self.internal_circles_nums):
            raise ValueError(
                f"{which_circle} is not one of the available internal circles")
        if where_in_that_circle<0 or where_in_that_circle>=self.internal_circles_nums[which_circle]:
            raise ValueError(
                f"{where_in_that_circle} is not available on internal circle {which_circle}")
        self.my_init_data[internal_bdry_vertex_name] = (BiColor.RED,[my_internal_connection])
        self.internal_bdry_orientations[which_circle][where_in_that_circle] = \
            internal_bdry_vertex_name
        self.is_set_internals[which_circle][where_in_that_circle] = True
        self.set_extras(internal_bdry_vertex_name,extras)

    def set_multi_edges(self,vertex_a : str, vertex_b: str, how_keys_match : Dict[int,int]) -> None:
        """
        set the permutation of half-edges if there are multiple edges connecting
        vertex_a and vertex_b
        """
        self.multi_edge_permutation[(vertex_a,vertex_b)] = how_keys_match

    def build(self) -> PlabicGraph:
        """
        use all the instance attributes in arguments to
        PlabicGraph construction
        """
        if not all(self.is_set_externals):
            raise ValueError("Not finished setting externals")
        if self.is_set_internals is not None and not all(all(z) for z in self.is_set_internals):
            raise ValueError("Not finished setting internals")
        for (vertex_a,vertex_b) in self.have_multi_edges:
            ab_in_multi = (vertex_a,vertex_b) in self.multi_edge_permutation
            ba_in_multi = (vertex_b,vertex_a) in self.multi_edge_permutation
            if not ab_in_multi and not ba_in_multi:
                raise ValueError("Not finished setting for multi-edges")
        return PlabicGraph(self.my_init_data,
                 self.external_init_orientation,
                 self.multi_edge_permutation,
                 self.internal_bdry_orientations,
                 self.extra_node_props)

# pylint:disable=too-many-public-methods,too-many-instance-attributes
class PlabicGraph:
    """
    built on top of a directed multigraph
    """

    num_interior_circles: int
    my_graph: nx.MultiDiGraph
    my_external_nodes: List[str]
    my_internal_bdry: List[List[str]]
    all_bdry_nodes: List[str]
    multi_edge_permutation: Dict[Tuple[str, str], Dict[int, int]]
    my_perfect_matching: Optional[Set[Tuple[str, str, int]]]
    my_extra_props : Set[str]
    circles_config : Optional[FramedDiskConfig]

    # pylint:disable = too-many-arguments, too-many-locals, too-many-branches,too-many-statements
    def __init__(self, my_init_data: Dict[str, Tuple[BiColor, List[str]]],
                 external_init_orientation: List[str],
                 multi_edge_permutation: Dict[Tuple[str, str], Dict[int, int]],
                 internal_bdry_orientations: Optional[List[List[str]]] = None,
                 extra_node_props: Optional[Dict[str, ExtraData]] = None,
                 my_circles_config : Optional[FramedDiskConfig] = None):
        """
        provide the data of the graph by a dictionary
        the keys are the names of the vertices
        the values are a tuple of it's color if it is internal
        and the clockwise list of it's neighboring nodes
        the list only matters up to rotation
        external_orientation is the list of boundary nodes and their indices
        give the 1 through n clockwise labels
        """
        if extra_node_props is None:
            extra_node_props = {}
        if internal_bdry_orientations is None:
            self.num_interior_circles = 0
            internal_bdry_orientations = []
        else:
            self.num_interior_circles = len(internal_bdry_orientations)
        self.my_graph = nx.MultiDiGraph()
        all_bdry_vertices = external_init_orientation.copy()
        for internal_circle_num in range(self.num_interior_circles):
            all_bdry_vertices.extend(
                internal_bdry_orientations[internal_circle_num])
        if len(set(all_bdry_vertices)) != len(all_bdry_vertices):
            raise ValueError("The boundary vertices must all be distinct")
        if any(z not in my_init_data for z in all_bdry_vertices):
            raise ValueError("The boundary vertices must all have data")
        self.my_extra_props : Set[str] = set()
        self.my_external_nodes = external_init_orientation
        self.my_internal_bdry = internal_bdry_orientations
        self.all_bdry_nodes = all_bdry_vertices
        for src, (cur_color, tgts) in my_init_data.items():
            is_cur_bdry = src in all_bdry_vertices
            if is_cur_bdry:
                self.my_graph.add_node(src, is_interior=False)
            else:
                self.my_graph.add_node(src, is_interior=True,
                                       color=cur_color)
            self.__add_props(src, extra_node_props.get(src, {}))
            if is_cur_bdry and len(tgts) != 1:
                raise ValueError(
                    "Each boundary vertex must be incident to a single edge")
            for edge_idx, tgt in enumerate(tgts):
                self.my_graph.add_edge(src, tgt, key=edge_idx)
        node_list = enumerate(my_init_data.keys())
        self.multi_edge_permutation: Dict[Tuple[str, str], Dict[int, int]] = {}
        any_self_loops = False
        for (src_num, src), (tgt_num, tgt) in itertools.combinations_with_replacement(node_list, 2):
            if src_num > tgt_num:
                continue
            if src_num == tgt_num:
                num_src_loops = self.num_connecting_edges(src, src)
                if num_src_loops == 0:
                    continue
                any_self_loops = True
                if num_src_loops > 1:
                    src_to_src_dict = multi_edge_permutation.get(
                        (src, src), None)
                    self.__multi_edge_permutation_add(
                        src, src, src_to_src_dict, None, num_src_loops)
                else:
                    msg = " ".join([f"If a self loop at {src},",
                                    "then there would be at least 2",
                                    "half edges there"])
                    raise ValueError(msg)
                continue
            num_src_tgt = self.num_connecting_edges(src, tgt)
            num_tgt_src = self.num_connecting_edges(tgt, src)
            if num_src_tgt != num_tgt_src:
                raise ValueError(
                    "\n".join([
                        "Each edge needs to have a corresponding backwards edge",
                        f"The one currently wrong was {src} {tgt}",
                        f"Forward {num_src_tgt}, Backwards {num_tgt_src}",
                        f"All out of {src} : {my_init_data[src][1]}",
                        f"All out of {tgt} : {my_init_data[tgt][1]}",
                        ]))
            if num_src_tgt > 1:
                src_to_tgt_dict = multi_edge_permutation.get((src, tgt), None)
                tgt_to_src_dict = multi_edge_permutation.get((tgt, src), None)
                self.__multi_edge_permutation_add(src, tgt, src_to_tgt_dict,
                                                  tgt_to_src_dict, num_src_tgt)
        self.__boundary_connectivity()
        self.my_perfect_matching: Optional[Set[Tuple[str, str, int]]] = set()
        self.__my_perfect_matching_fix(any_self_loops)

        self.circles_config = None
        if my_circles_config is not None:
            self.circles_config = my_circles_config
            # raise NotImplementedError(
            #    "are all the boundary vertices on the circles they are supposed to be")

    def __add_props(self, vertex: str, props: ExtraData):
        """
        add all the extra keys present for extra data to attach to nodes
        for example, position
        """
        for prop_name, prop_val in props.items():
            if prop_name not in ["is_interior", "color"]:
                self.my_graph.nodes[vertex][prop_name] = prop_val
                self.my_extra_props.add(prop_name)

    def remove_prop(self, prop_name : str):
        """
        remove a property from all nodes
        """
        if prop_name in ["is_interior", "color"]:
            raise ValueError("Can only delete the optional properties")
        for cur_node in self.my_graph.nodes:
            try:
                del self.my_graph.nodes[cur_node][prop_name]
            except KeyError:
                pass
        self.my_extra_props.discard(prop_name)

    def __boundary_connectivity(self):
        """
        make sure every internal vertex is connected to some external boundary vertex
        """
        infinity_node = "infinity"
        assert infinity_node not in self.my_graph
        self.my_graph.add_node(
            infinity_node, is_interior=False, color=BiColor.RED)
        for edge_idx, bdry in enumerate(self.my_external_nodes):
            self.my_graph.add_edge(infinity_node, bdry, edge_idx)
        if not nx.is_weakly_connected(self.my_graph):
            self.my_graph.remove_node(infinity_node)
            raise ValueError(
                "Every internal vertex should be connected to some boundary vertex")
        self.my_graph.remove_node(infinity_node)

    def __clear_perfect_matching(self):
        """
        clears everything to do with perfect orientation
        """
        self.remove_prop("my_perfect_edge")
        self.my_perfect_matching = None

    def __my_perfect_matching_fix(self, any_self_loops: bool):
        """
        makes sure the numbers giving my_perfect_edge
        on each vertex induce a perfect matching
        possibly after ignoring some of the boundary vertices
        """
        if any_self_loops:
            self.__clear_perfect_matching()
            return
        if not self.is_bipartite():
            self.__clear_perfect_matching()
            return
        all_node_names = list(self.my_graph.nodes())
        try:
            special_edge_numbers = {
                z: self.my_graph.nodes[z]["my_perfect_edge"] for z in all_node_names}
        except KeyError:
            self.__clear_perfect_matching()
            return
        self.my_perfect_matching: Optional[Set[Tuple[str, str, int]]] = set()
        already_partnered: Dict[str, str] = {}
        for src_name, tgt_name, edge_key in self.my_graph.edges(keys=True):
            if edge_key == special_edge_numbers[src_name]:
                if src_name in already_partnered:
                    raise ValueError(
                        f"Should not have seen {src_name} as a source yet")
                already_partnered[src_name] = tgt_name
                if already_partnered.get(tgt_name, src_name) != src_name:
                    self.__clear_perfect_matching()
                    return
                self.my_perfect_matching.add((src_name, tgt_name, edge_key))
        for src_name in self.my_graph:
            if src_name not in already_partnered and \
                    self.my_graph.nodes[src_name]["is_interior"]:
                self.__clear_perfect_matching()
                break

    def __multi_edge_permutation_add(self, src: str, tgt: str,
                                     src_to_tgt_dict: Optional[Dict[int, int]],
                                     tgt_to_src_dict: Optional[Dict[int, int]],
                                     num_edges: int):
        """
        helper for the initializer/add_nodes with
        the permutation of multi-edge indices from
        the permutation of the half edge numbers
        """
        if num_edges <= 1:
            return
        if src_to_tgt_dict is None and tgt_to_src_dict is None:
            msg = "".join(["When there are multi-edges, a dictionary mapping the edge numbers",
                           f"of the {src} to {tgt} half edges to the {tgt} to {src} half edges"])
            raise ValueError(msg)
        if src_to_tgt_dict is not None and tgt_to_src_dict is not None:
            msg = "".join(["When there are multi-edges,",
                           " exactly one dictionary mapping the edge numbers",
                           f"of the {src} to {tgt} half edges to the {tgt} to {src} half edges",
                           "must be provided. The other direction is the inverse."])
            raise ValueError(msg)

        def make_perm(src_temp: str, tgt_temp: str, which_dict: Dict[int, int]) -> Dict[int, int]:
            """
            check validity of the permutation of the half edge numbers
            which indicates the cyclic orderings
            """
            src_temp_tgt_temp_edge_numbers: List[int] = \
                list(self.my_graph[src_temp][tgt_temp].keys())
            tgt_temp_src_temp_edge_numbers: List[int] = \
                list(self.my_graph[tgt_temp][src_temp].keys())
            if set(which_dict.keys()) != set(src_temp_tgt_temp_edge_numbers):
                msg = " ".join(["When there are multi-edges,",
                               "need a dictionary mapping the edge numbers",
                                f"of the {src} to {tgt} half edges to",
                                "the {tgt} to {src} half edges"])
                raise ValueError(msg)
            if set(which_dict.values()) != tgt_temp_src_temp_edge_numbers:
                msg = " ".join(["When there are multi-edges,",
                               "need a dictionary mapping the edge numbers",
                                f"of the {src} to {tgt} half edges to",
                                "the {tgt} to {src} half edges"])
                raise ValueError(msg)
            return which_dict
        if src_to_tgt_dict is None:
            self.multi_edge_permutation[(tgt, src)] = \
                make_perm(tgt, src, cast(Dict[int, int], tgt_to_src_dict))
        else:
            self.multi_edge_permutation[(src, tgt)] = \
                make_perm(src, tgt, cast(Dict[int, int], src_to_tgt_dict))

    def is_trivalent(self, this_vertex: str) -> bool:
        """
        is this_vertex trivalent
        """
        return len(self.my_graph[this_vertex]) == 3

    def one_color_trivalent(self, all_of_this_color: BiColor) -> bool:
        """
        are all vertices of specified color trivalent
        """
        return all(
            self.get_color(z) != all_of_this_color or self.is_trivalent(z)
            for z in self.my_graph)

    def boundary_connects_to_color(self, all_of_this_color: BiColor) -> bool:
        """
        are boundary vertices connected to only internal vertices of specified color
        no boundary to boundary allowed, no boundary to the opposite color allowed
        """
        for cur_bdry in self.all_bdry_nodes:
            for neighbor in self.my_graph.neighbors(cur_bdry):
                if self.get_color(neighbor) != all_of_this_color:
                    return False
        return True

    def is_bipartite(self) -> bool:
        """
        are all edges connecting internal vertices opposite colors
        """
        for src, tgt in itertools.combinations(self.my_graph, 2):
            if not self.my_graph.nodes[src]["is_interior"] or \
                    not self.my_graph.nodes[tgt]["is_interior"]:
                continue
            if self.nodes_connected(src, tgt) and not self.opposite_colors(src, tgt):
                return False
        return True

    def nodes_connected(self, this_vertex: str, that_vertex: str) -> bool:
        """
        are this_vertex and that_vertex connected by an edge
        """
        num_edges = self.num_connecting_edges(this_vertex, that_vertex)
        return num_edges > 0

    def num_connecting_edges(self, this_vertex: str, that_vertex: str) -> int:
        """
        how many edges connect this_vertex and that_vertex
        """
        try:
            num_edges_connecting = len(self.my_graph[this_vertex][that_vertex])
        except KeyError:
            num_edges_connecting = 0
        return num_edges_connecting

    def __corresponding_multiedge_idx(self, src: str, tgt: str, key_in_src_tgt: int) -> int:
        """
        if we are looking at the half edge labelled key_in_src_tgt'th connecting src to tgt
        then what is the key of the corresponding half edge backwards from tgt to src
        """
        assert 0 <= key_in_src_tgt < self.my_graph.out_degree(src)
        num_edges = self.num_connecting_edges(src, tgt)
        if num_edges == 1:
            all_keys = list(self.my_graph[tgt][src].keys())
            return all_keys[0]
        if num_edges == 0:
            raise ValueError(" ".join([f"There should have been edges from {src} to {tgt}",
                             f"so we could get the one with key {key_in_src_tgt}"]))
        perm = self.multi_edge_permutation.get((src, tgt), None)
        if perm is not None:
            return perm[key_in_src_tgt]
        perm_inv = self.multi_edge_permutation.get((tgt, src), None)
        assert perm_inv is not None
        perm = {}
        to_return = None
        for (key_tgt_src, key_src_tgt) in perm_inv.items():
            if key_src_tgt == key_in_src_tgt:
                to_return = key_tgt_src
            perm[key_src_tgt] = key_tgt_src
        self.multi_edge_permutation[(src, tgt)] = perm
        return cast(int, to_return)

    def get_color(self, this_vertex: str) -> Optional[BiColor]:
        """
        the color of this vertex
        """
        return cast(Optional[BiColor], self.my_graph.nodes()[this_vertex].get("color", None))

    def opposite_colors(self, this_vertex: str, that_vertex: str) -> bool:
        """
        are this_vertex and that_vertex both colored and opposite
        """
        this_color = self.get_color(this_vertex)
        that_color = self.get_color(that_vertex)
        if this_color is None or that_color is None:
            return False
        return this_color != that_color

    def _by_edge_number(self, src: str, desired_edge_number: int) -> Tuple[str, int]:
        """
        the target of the half-edge starting at src
        with the desired edge number in the cyclic ordering
        also the corresponding half-edge when looking from target to src
        """
        for tgt in self.my_graph[src]:
            if desired_edge_number in self.my_graph[src][tgt]:
                flipped = self.__corresponding_multiedge_idx(
                    src, tgt, desired_edge_number)
                return tgt, flipped
        raise ValueError(
            f"Nothing with edge number {desired_edge_number} out of {src} found")

    def greedy_shrink(self,
                      name_combiner : Optional[Callable[[str,str],str]] = None,
                      rounds_left : int = 5) -> bool:
        """
        do as many M2 and M3 that remove vertices or edges
        provided no extra data is present
        the name_combiner tells what to do with names on the
        contract edge moves
        """
        if rounds_left>=5 and not self.my_extra_props.issubset([]):
            return False
        name_combiner = name_combiner if name_combiner is not None else lambda z1,_: z1
        any_change = False
        all_edges = list(self.my_graph.edges(keys=False))
        for name1,name2 in all_edges:
            if name1 in self.my_graph.nodes() and name2 in self.my_graph.nodes():
                did_remove,_ = self.contract_edge(name1,name2,name_combiner(name1,name2))
            else:
                did_remove = False
            any_change = any_change or did_remove
        all_node_names = list(self.my_graph.nodes())
        for name in all_node_names:
            did_remove,_ = self.remove_bivalent(name)
            any_change = any_change or did_remove
        if not any_change:
            return False
        if rounds_left>0:
            _ = self.greedy_shrink(name_combiner,rounds_left-1)
        return True

    def square_move(self, four_nodes: Tuple[str, str, str, str]) -> Tuple[bool, str]:
        """
        perform a square move on the quadrilateral with these vertices
        if these names did not meet the pre-conditions for a square move
            provide the explanation as well
        also output if self has changed
        """
        if len(set(four_nodes)) != 4:
            return False, "The 4 need to be distinct"
        all_trivalent = all((self.is_trivalent(cur_node)
                            for cur_node in four_nodes))
        if not all_trivalent:
            return False, "The 4 must be trivalent"
        is_quad = True
        alternating_colors = True

        for cur_node, next_node in zip(four_nodes, four_nodes[1:]):
            is_quad = is_quad and self.nodes_connected(cur_node, next_node)
            alternating_colors = alternating_colors and self.opposite_colors(
                cur_node, next_node)
        is_quad = is_quad and self.nodes_connected(
            four_nodes[3], four_nodes[0])
        alternating_colors = alternating_colors and\
            self.opposite_colors(four_nodes[3], four_nodes[0])
        if not is_quad:
            return False, "The 4 must form a quadrilateral"
        if not alternating_colors:
            return False, "The 4 must alternate colors"

        for cur_node in four_nodes:
            cur_color = self.get_color(cur_node)
            if cur_color is BiColor.RED:
                self.my_graph.nodes()[cur_node]["color"] = BiColor.GREEN
            else:
                self.my_graph.nodes()[cur_node]["color"] = BiColor.RED

        # clears everything to do with perfect orientation
        # could fix the perfect orientation instead
        # along this move
        self.__clear_perfect_matching()

        return True, "Success"

    # pylint:disable = too-many-return-statements, too-many-statements
    def flip_move(self, this_vertex: str,
                  that_vertex: str,
                  extra_data_transformer:
                  Optional[Callable[[ExtraData, ExtraData, "PlabicGraph"],
                                    Tuple[ExtraData, ExtraData]]] = None)\
            -> Tuple[bool, str]:
        """
        perform a flip move on the edge with these vertices
        if these names did not meet the pre-conditions for a square move
            provide the explanation as well
        also output if self has changed
        """
        if this_vertex == that_vertex:
            return False, "The 2 need to be distinct"
        if not self.is_trivalent(this_vertex) and self.is_trivalent(that_vertex):
            return False, "The 2 must be trivalent"
        if self.num_connecting_edges(this_vertex, that_vertex) != 1:
            msg = "\n".join(["The 2 must have exactly 1 edge connecting them.",
                             "If there were multiple, we don't know which of them",
                             "to collapse then re-expand"])
            return False, msg
        if self.num_connecting_edges(this_vertex, this_vertex) > 0:
            return False, "No self loops for this implementation of flip move to work"
        if self.num_connecting_edges(that_vertex, that_vertex) > 0:
            return False, "No self loops for this implementation of flip move to work"
        first_color = self.get_color(this_vertex)
        second_color = self.get_color(that_vertex)
        if first_color is None or second_color is None or first_color != second_color:
            return False, "The two must be the same color"
        cut_edge_number_this = list(
            self.my_graph[this_vertex][that_vertex].keys())[0]
        cut_edge_number_that = list(
            self.my_graph[that_vertex][this_vertex].keys())[0]

        # if starting from the left hand side picture
        #   if this is the top and that is the bottom
        #       ABCD is upper_left, upper_right, lower_left, lower_right
        #   if this is the bottom and that is the top
        #       ABCD is lower_right, lower_left, upper_right, upper_left
        # if starting from the right hand side picture
        #   if this is the left and that is the right
        #       ABCD is lower_left,upper_left,lower_right,upper_right
        #   if this is the right and that is the left
        #       ABCD is upper_right,lower_right,upper_left,lower_left

        edge_number_a = cut_edge_number_this + 1
        if edge_number_a >= 3:
            edge_number_a -= 3
        vertex_a, halfedge_number_from_a = \
            self._by_edge_number(this_vertex, edge_number_a)

        edge_number_b = cut_edge_number_this - 1
        if edge_number_b < 0:
            edge_number_b += 3
        vertex_b, halfedge_number_from_b = \
            self._by_edge_number(this_vertex, edge_number_b)

        edge_number_d = cut_edge_number_that + 1
        if edge_number_d >= 3:
            edge_number_d -= 3
        vertex_d, halfedge_number_from_d = \
            self._by_edge_number(that_vertex, edge_number_d)

        edge_number_c = cut_edge_number_that - 1
        if edge_number_c < 0:
            edge_number_c += 3
        vertex_c, halfedge_number_from_c = \
            self._by_edge_number(that_vertex, edge_number_c)

        for node_to_go in [this_vertex, that_vertex]:
            adjacencies = list(self.my_graph[node_to_go])
            for adj in adjacencies:
                if self.multi_edge_permutation.get((node_to_go, adj), None) is not None:
                    del self.multi_edge_permutation[(node_to_go, adj)]
                if self.multi_edge_permutation.get((adj, node_to_go), None) is not None:
                    del self.multi_edge_permutation[(adj, node_to_go)]
        if extra_data_transformer is not None:
            old_this_data = cast(ExtraData, self.my_graph.nodes[this_vertex])
            old_that_data = cast(ExtraData, self.my_graph.nodes[that_vertex])
            new_this_data, new_that_data = extra_data_transformer(
                old_this_data, old_that_data, self)
        self.my_graph.remove_node(this_vertex)
        self.my_graph.remove_node(that_vertex)

        self.my_graph.add_node(this_vertex,
                               is_interior=True,
                               color=first_color)
        self.my_graph.add_node(that_vertex,
                               is_interior=True,
                               color=second_color)
        if extra_data_transformer is not None:
            self.__add_props(this_vertex, new_this_data)
            self.__add_props(that_vertex, new_that_data)

        self.my_graph.add_edge(vertex_a, this_vertex, halfedge_number_from_a)
        self.my_graph.add_edge(this_vertex, vertex_a, 2)

        self.my_graph.add_edge(that_vertex, this_vertex, 2)
        self.my_graph.add_edge(this_vertex, that_vertex, 0)

        self.my_graph.add_edge(vertex_c, this_vertex, halfedge_number_from_c)
        self.my_graph.add_edge(this_vertex, vertex_c, 1)

        self.my_graph.add_edge(vertex_b, that_vertex, halfedge_number_from_b)
        self.my_graph.add_edge(that_vertex, vertex_b, 0)

        self.my_graph.add_edge(vertex_d, that_vertex, halfedge_number_from_d)
        self.my_graph.add_edge(that_vertex, vertex_d, 1)

        if vertex_a == vertex_c:
            self.__multi_edge_permutation_add(
                this_vertex, vertex_a,
                {2: halfedge_number_from_a,
                 1: halfedge_number_from_c}, None, 2)
        if vertex_b == vertex_d:
            self.__multi_edge_permutation_add(
                that_vertex, vertex_b,
                {0: halfedge_number_from_b,
                 1: halfedge_number_from_d}, None, 2)

        # clears everything to do with perfect orientation
        # could fix the perfect orientation instead
        # along this move
        self.__clear_perfect_matching()

        return True, "Success"

    def remove_bivalent(self, my_bivalent_vertex: str) -> Tuple[bool, str]:
        """
        remove the specified bivalent vertex
        if these names did not meet the pre-conditions of being bivalent
            provide the explanation as well
        also output if self has changed
        """
        should_be_2 = self.my_graph.out_degree(my_bivalent_vertex)
        if should_be_2 != 2:
            return False, f"{my_bivalent_vertex} needs to be bivalent"
        side_1, idx_from_side_1 = self._by_edge_number(my_bivalent_vertex, 0)
        if side_1 == my_bivalent_vertex:
            return False, "No self loops for this implementation of remove bivalent to work"
        side_2, idx_from_side_2 = self._by_edge_number(my_bivalent_vertex, 1)
        if side_2 == my_bivalent_vertex:
            return False, "No self loops for this implementation of remove bivalent to work"
        if self.get_color(side_1) is None and self.get_color(side_2) is None:
            return False, "Cannot cause two boundary nodes to connect directly"
        num_side_1_to_side_2 = self.num_connecting_edges(side_1, side_2)
        if num_side_1_to_side_2 > 1:
            old_dict_12 = self.multi_edge_permutation.get(
                (side_1, side_2), None)
            old_dict_21 = self.multi_edge_permutation.get(
                (side_2, side_1), None)
        if num_side_1_to_side_2 == 1:
            all_keys_12 = list(self.my_graph[side_1][side_2].keys())
            all_keys_21 = list(self.my_graph[side_2][side_1].keys())
            old_dict_12 = {all_keys_12[0]: all_keys_21[0]}
            old_dict_21 = None
        if num_side_1_to_side_2 == 0:
            old_dict_12 = None
            old_dict_21 = None
        for adj in [side_1, side_2]:
            if self.multi_edge_permutation.get((my_bivalent_vertex, adj), None) is not None:
                del self.multi_edge_permutation[(my_bivalent_vertex, adj)]
            if self.multi_edge_permutation.get((adj, my_bivalent_vertex), None) is not None:
                del self.multi_edge_permutation[(adj, my_bivalent_vertex)]
        self.my_graph.remove_node(my_bivalent_vertex)
        self.my_graph.add_edge(side_1, side_2, idx_from_side_1)
        self.my_graph.add_edge(side_2, side_1, idx_from_side_2)
        if old_dict_12 is not None:
            old_dict_12[idx_from_side_1] = idx_from_side_2
            self.multi_edge_permutation[(side_1, side_2)] = old_dict_12
        if old_dict_21 is not None:
            old_dict_21[idx_from_side_2] = idx_from_side_1
            self.multi_edge_permutation[(side_2, side_1)] = old_dict_21

        # clears everything to do with perfect orientation
        # could fix the perfect orientation instead
        # along this move
        self.__clear_perfect_matching()

        return True, "Success"

    def insert_bivalent(self, this_vertex: str, that_vertex: str,
                        desired_color: BiColor,
                        desired_name: str,
                        extra_data_transformer:
                            Optional[Callable[[ExtraData, ExtraData, "PlabicGraph"],
                                              ExtraData]] = None) -> Tuple[bool, str]:
        """
        add a bivalent vertex on the edge connecting these two vertices
        if these names did not meet the pre-conditions of having an edge
            provide the explanation as well
        also output if self has changed
        """
        num_edges_connecting = self.num_connecting_edges(
            this_vertex, that_vertex)
        if num_edges_connecting != 1:
            return False, "There needs to be exactly one edge to split specified by these vertices"
        if desired_name in self.my_graph.nodes():
            return False, "The new vertex name should be distinct from those present"
        key_this_that_all: List[int] = list(
            self.my_graph[this_vertex][that_vertex].keys())
        key_this_that = key_this_that_all[0]
        key_that_this = self.__corresponding_multiedge_idx(
            this_vertex, that_vertex, key_this_that)

        if extra_data_transformer is not None:
            old_this_data = cast(ExtraData, self.my_graph.nodes[this_vertex])
            old_that_data = cast(ExtraData, self.my_graph.nodes[that_vertex])
            new_props = extra_data_transformer(
                old_this_data, old_that_data, self)
        self.my_graph.add_node(
            desired_name, is_interior=True, color=desired_color)
        if extra_data_transformer is not None:
            self.__add_props(desired_name, new_props)
        self.my_graph.remove_edge(this_vertex, that_vertex)
        self.my_graph.add_edge(this_vertex, desired_name, key_this_that)
        self.my_graph.add_edge(desired_name, this_vertex, 0)
        self.my_graph.add_edge(that_vertex, desired_name, key_that_this)
        self.my_graph.add_edge(desired_name, that_vertex, 1)

        # clears everything to do with perfect orientation
        # could fix the perfect orientation instead
        # along this move
        self.__clear_perfect_matching()

        return True, "Success"

    def contract_edge(self, this_vertex: str,
                      that_vertex: str,
                      combined_name: str,
                      extra_data_transformer:
                      Optional[Callable[[ExtraData, ExtraData, "PlabicGraph"],
                                        ExtraData]] = None)\
            -> Tuple[bool, str]:
        """
        contract the edge connecting these two vertices of the same color
        if these names did not meet the pre-conditions
            provide the explanation as well
        also output if self has changed
        """
        if this_vertex == that_vertex:
            return False, "No self loops for this implementation of contract edge to work"
        this_color = self.get_color(this_vertex)
        that_color = self.get_color(that_vertex)
        if this_color is None or that_color is None:
            return False, "The two vertices for the collapsing edge must be internal"
        if this_color != that_color:
            return False, "They must be the same color, to collapse into 1"
        num_edges_connecting = self.num_connecting_edges(
            this_vertex, that_vertex)
        if num_edges_connecting != 1:
            return False,\
                "There needs to be exactly one edge to collapse specified by these vertices"
        if combined_name in self.my_graph.nodes() and \
                combined_name not in [this_vertex, that_vertex]:
            return False, "The new vertex name should be distinct from those present"
        all_adjacencies: List[str] = list(self.my_graph[this_vertex])
        all_adjacencies.extend(self.my_graph[that_vertex])
        connects_to_this = [False] * len(all_adjacencies)
        connects_to_that = [False] * len(all_adjacencies)
        will_multiconnect = []
        for idx, cur_adj in enumerate(all_adjacencies):
            num_to_this = self.num_connecting_edges(cur_adj, this_vertex)
            connects_to_this[idx] = num_to_this > 0
            num_to_that = self.num_connecting_edges(cur_adj, that_vertex)
            connects_to_that[idx] = num_to_that > 0
            if num_to_this+num_to_that > 1:
                will_multiconnect.append(cur_adj)
        key_this_that_all: List[int] = list(
            self.my_graph[this_vertex][that_vertex].keys())
        key_this_that = key_this_that_all[0]
        out_this = self.my_graph.out_degree(this_vertex)
        key_that_this = self.__corresponding_multiedge_idx(
            this_vertex, that_vertex, key_this_that)
        out_that = self.my_graph.out_degree(that_vertex)
        order_around_combined: List[Tuple[bool, int, Tuple[str, int]]] = []
        order_around_combined.extend(((True, z, self._by_edge_number(this_vertex, z))
                                      for z in range(key_this_that+1, out_this)))
        order_around_combined.extend(((True, z, self._by_edge_number(this_vertex, z))
                                      for z in range(0, key_this_that)))
        order_around_combined.extend(((False, z, self._by_edge_number(that_vertex, z))
                                      for z in range(key_that_this+1, out_that)))
        order_around_combined.extend(((False, z, self._by_edge_number(that_vertex, z))
                                      for z in range(0, key_that_this)))

        for node_to_go in [this_vertex, that_vertex]:
            adjacencies = list(self.my_graph[node_to_go])
            for adj in adjacencies:
                if self.multi_edge_permutation.get((node_to_go, adj), None) is not None:
                    del self.multi_edge_permutation[(node_to_go, adj)]
                if self.multi_edge_permutation.get((adj, node_to_go), None) is not None:
                    del self.multi_edge_permutation[(adj, node_to_go)]
        if extra_data_transformer is not None:
            old_this_data = cast(ExtraData, self.my_graph.nodes[this_vertex])
            old_that_data = cast(ExtraData, self.my_graph.nodes[that_vertex])
            new_props = extra_data_transformer(
                old_this_data, old_that_data, self)
        self.my_graph.remove_node(this_vertex)
        self.my_graph.remove_node(that_vertex)

        self.my_graph.add_node(
            combined_name, is_interior=True, color=this_color)
        if extra_data_transformer is not None:
            self.__add_props(combined_name, new_props)
        for key_from_combined, (_origin_this, _origin_key,
                                (tgt, half_edge_from_tgt)) in\
                enumerate(order_around_combined):
            self.my_graph.add_edge(combined_name, tgt, key_from_combined)
            self.my_graph.add_edge(tgt, combined_name,
                                   half_edge_from_tgt)
            if tgt in will_multiconnect:
                if (combined_name, tgt) in self.multi_edge_permutation:
                    self.multi_edge_permutation[(combined_name, tgt)][key_from_combined] =\
                        half_edge_from_tgt
                else:
                    self.multi_edge_permutation[(combined_name, tgt)] = {}
                    self.multi_edge_permutation[(combined_name, tgt)][key_from_combined] =\
                        half_edge_from_tgt

        # clears everything to do with perfect orientation
        # could fix the perfect orientation instead
        # along this move
        self.__clear_perfect_matching()

        return True, "Success"

    def split_vertex(self, this_vertex: str,
                     split_bounds: Tuple[int, int],
                     split_name_1: str,
                     split_name_2: str,
                     extra_data_transformer:
                     Optional[Callable[[ExtraData, "PlabicGraph"],
                                       Tuple[ExtraData, ExtraData]]] = None)\
            -> Tuple[bool, str]:
        """
        split this vertex into an edge connecting two vertices of the same color
        split_bounds informs which edges attach to which
            it splits the cyclic ordering into two linear orderings
        if anything did not meet the pre-conditions
            provide the explanation as well
        also output if self has changed
        """
        this_color = self.get_color(this_vertex)
        if this_color is None:
            return False, "The vertex for splitting must be internal"
        if split_name_1 == split_name_2:
            return False, "The new vertex names should be distinct from each other"
        if split_name_1 in self.my_graph.nodes() or \
                split_name_2 in self.my_graph.nodes():
            if split_name_1 == this_vertex and split_name_2 not in self.my_graph.nodes():
                pass
            elif split_name_2 == this_vertex and split_name_1 not in self.my_graph.nodes():
                pass
            else:
                return False, "The new vertex names should be distinct from those present"
        this_out_degree = self.my_graph.out_degree(this_vertex)
        if split_bounds[0] <= split_bounds[1]:
            go_to_1 = list(range(split_bounds[0], split_bounds[1]+1))
            go_to_2 = list(range(split_bounds[1]+1, this_out_degree)) +\
                list(range(0, split_bounds[0]))
        else:
            go_to_2 = list(range(split_bounds[1], split_bounds[0]+1))
            go_to_1 = list(range(split_bounds[0]+1, this_out_degree)) +\
                list(range(0, split_bounds[1]))

        go_to_1_full = []
        seen_on_1 = set()
        will_multiconnect_1 = set()
        for on_1_key, on_this_key in enumerate(go_to_1):
            tgt, on_tgt_key = self._by_edge_number(this_vertex, on_this_key)
            go_to_1_full.append((on_1_key, tgt, on_tgt_key))
            if tgt in seen_on_1:
                will_multiconnect_1.add(tgt)
            seen_on_1.add(tgt)
        go_to_2_full = []
        seen_on_2 = set()
        will_multiconnect_2 = set()
        for on_2_key, on_this_key in enumerate(go_to_2):
            tgt, on_tgt_key = self._by_edge_number(this_vertex, on_this_key)
            go_to_2_full.append((on_2_key, tgt, on_tgt_key))
            if tgt in seen_on_2:
                will_multiconnect_2.add(tgt)
            seen_on_2.add(tgt)

        adjacencies = list(self.my_graph[this_vertex])
        for adj in adjacencies:
            if self.multi_edge_permutation.get((this_vertex, adj), None) is not None:
                del self.multi_edge_permutation[(this_vertex, adj)]
            if self.multi_edge_permutation.get((adj, this_vertex), None) is not None:
                del self.multi_edge_permutation[(adj, this_vertex)]
        if extra_data_transformer is not None:
            old_this_data = cast(ExtraData, self.my_graph.nodes[this_vertex])
            new_props_1, new_props_2 = extra_data_transformer(
                old_this_data, self)
        self.my_graph.remove_node(this_vertex)

        self.my_graph.add_node(
            split_name_1, is_interior=True, color=this_color)
        self.my_graph.add_node(
            split_name_2, is_interior=True, color=this_color)
        if extra_data_transformer is not None:
            self.__add_props(split_name_1, new_props_1)
            self.__add_props(split_name_2, new_props_2)
        out_on_1 = len(go_to_1_full)
        out_on_2 = len(go_to_2_full)
        self.my_graph.add_edge(split_name_1, split_name_2, out_on_1)
        self.my_graph.add_edge(split_name_2, split_name_1, out_on_2)
        for on_1_key, tgt, on_tgt_key in go_to_1_full:
            self.my_graph.add_edge(split_name_1, tgt, on_1_key)
            self.my_graph.add_edge(tgt, split_name_1, on_tgt_key)
            if tgt in will_multiconnect_1:
                if (split_name_1, tgt) not in self.multi_edge_permutation:
                    self.multi_edge_permutation[(split_name_1, tgt)] = {}
                self.multi_edge_permutation[(
                    split_name_1, tgt)][on_1_key] = on_tgt_key
        for on_2_key, tgt, on_tgt_key in go_to_2_full:
            self.my_graph.add_edge(split_name_2, tgt, on_2_key)
            self.my_graph.add_edge(tgt, split_name_2, on_tgt_key)
            if tgt in will_multiconnect_2:
                if (split_name_2, tgt) not in self.multi_edge_permutation:
                    self.multi_edge_permutation[(split_name_2, tgt)] = {}
                self.multi_edge_permutation[(
                    split_name_2, tgt)][on_2_key] = on_tgt_key

        # clears everything to do with perfect orientation
        # could fix the perfect orientation instead
        # along this move
        self.__clear_perfect_matching()

        return True, "Success"

    def __operad_compose_helper(self,which_internal_disk : int,
                                   other:PlabicGraph,
                                   glued_vertices : List[str],
                                   force_it : bool = False)\
                                    -> Tuple[bool,Callable[[Point],Point]]:
        """
        find a transformation
        taking the positions of the glued vertices in self
        to the corresponding positions in other
        does that transformation to self.circles_config
        """
        try:
            all_positions_self = {
                    z: self.my_graph.nodes[z]["position"] for z in glued_vertices}
            all_positions_other = {
                    z: other.my_graph.nodes[z]["position"] for z in glued_vertices}
        except KeyError:
            # something didn't have position so can't trust circles_config
            # or positions in the operad composition
            # circles_config is removed here
            # the False output tells operad_compose to clear position property
            self.circles_config = None
            return False,lambda z: z
        if self.circles_config is not None and other.circles_config is not None:
            supposed_transformation = \
                self.circles_config.do_transformation(which_internal_disk,other.circles_config)
            transformed_self = {z1:supposed_transformation(cast(Point,z2))
                                for (z1,z2) in all_positions_self.items()}
            if transformed_self == all_positions_other:
                # the transformations of the framed circles accomplished the task
                # the circles_config has already changed
                # the points will all change later in operad_compose
                return True,supposed_transformation
            # the transformations of the framed circles failed the task
            # circles_config is removed here
            # the False output tells operad_compose to clear position property
            self.circles_config = None
            return False,lambda z: z
        if all_positions_self == all_positions_other:
            # the points lined up but we didn't have both circles_config for self and other
            # so we can't have one in the composition
            # but we still have the positions of the vertices in the operad composition
            self.circles_config = None
            return True,lambda z: z
        if not force_it:
            # the points did not line up and we didn't have both circles_config for self and other
            # so we can't have one in the composition
            # and we aren't forcing to find an affine transformation that does the task without
            # the help of FramedDisKConfig
            self.circles_config = None
            return False,lambda z: z
        raise NotImplementedError

    def coordinate_transform(self,transform : Callable[[Point],Point],
                             circle_config_invalidates : bool = True) -> bool:
        """
        does transform on all positions of all nodes
        """
        def point_like(maybe_point) -> bool:
            """
            is p in R^2
            """
            return isinstance(maybe_point,tuple) and len(maybe_point)==2 and\
                isinstance(maybe_point[0],numbers.Real) and isinstance(maybe_point[1],numbers.Real)
        all_are_points = True
        for node_name in self.my_graph.nodes():
            old_position = self.my_graph.nodes[node_name].get("position",None)
            if not point_like(old_position):
                # print(f"{node_name} gave {old_position}")
                all_are_points = False
                break
        if all_are_points:
            if circle_config_invalidates:
                self.circles_config = None
            for node_name in self.my_graph.nodes():
                old_position = self.my_graph.nodes[node_name]["position"]
                new_position = transform(old_position)
                self.my_graph.nodes[node_name]["position"] = new_position
        return all_are_points

    def operad_compose(self, other: PlabicGraph, which_internal_disk: int) -> Tuple[bool, str]:
        """
        substitute other in on the i'th internal circle of self
            that is self has at least that many in self.my_internal_bdry
            and self.my_internal_bdry[which_internal_disk] are matched up
            with the other.my_external_nodes
        if anything did not meet the pre-conditions
            provide the explanation as well
        also output if self has changed
        """
        if which_internal_disk < 0:
            return False, "Which internal disk is a natural number"
        if self.num_interior_circles <= which_internal_disk:
            return False, "Not enough internal disks"
        relevant_internal = self.my_internal_bdry[which_internal_disk]

        def cyclic_equal(a_list: List[str], b_list: List[str]) -> bool:
            """
            some cyclic permutation of a_list is equal to b_list
            """
            if len(a_list) != len(b_list):
                return False
            for shift in range(len(a_list)):
                up_to = len(b_list) - shift
                first_part = a_list[shift:len(a_list)] == b_list[0:up_to]
                second_part = a_list[0:shift] == b_list[up_to:len(b_list)]
                if first_part and second_part:
                    return True
            return False

        if not cyclic_equal(relevant_internal, other.my_external_nodes):
            return False, "The boundary vertices that should be glued did not match up"

        nodes_overlapping = set(self.my_graph.nodes())
        nodes_overlapping = nodes_overlapping.intersection(
            set(other.my_graph.nodes()))
        if nodes_overlapping != set(relevant_internal):
            return False, "The only overlap of vertex names should be the ones that are glued"

        for replaced_node in relevant_internal:
            self_tgt, _ = self._by_edge_number(replaced_node, 0)
            # pylint:disable = protected-access
            other_tgt, _ = other._by_edge_number(replaced_node, 0)
            if self.get_color(self_tgt) is None or \
                    other.get_color(other_tgt) is None:
                return False,\
                    "Boundary vertices should connect to internal vertices not directly to boundary"
        if "position" in self.my_extra_props and "position" in other.my_extra_props:
            found_transformation, my_transform = \
                self.__operad_compose_helper(which_internal_disk, other,relevant_internal)
            clear_position = not found_transformation
            if found_transformation:
                self.coordinate_transform(my_transform,False)
        else:
            clear_position = True

        self.all_bdry_nodes = self.my_external_nodes
        new_internal_bdry: List[List[str]] = \
            [[]]*(self.num_interior_circles-1+other.num_interior_circles)
        for idx_l in range(which_internal_disk):
            new_internal_bdry[idx_l] = self.my_internal_bdry[idx_l]
            self.all_bdry_nodes.extend(new_internal_bdry[idx_l])
        idx_l = which_internal_disk
        for idx_r in range(other.num_interior_circles):
            new_internal_bdry[idx_l+idx_r] = other.my_internal_bdry[idx_r]
            self.all_bdry_nodes.extend(new_internal_bdry[idx_l+idx_r])
        idx_l = other.num_interior_circles
        for idx_r in range(which_internal_disk, self.num_interior_circles):
            new_internal_bdry[idx_l+idx_r] = self.my_internal_bdry[idx_r]
            self.all_bdry_nodes.extend(new_internal_bdry[idx_l+idx_r])
        self.my_internal_bdry = new_internal_bdry
        self.num_interior_circles += other.num_interior_circles - 1

        for replaced_node in relevant_internal:
            tgt, _ = self._by_edge_number(replaced_node, 0)
            self.my_graph.add_edge(replaced_node, tgt, 1)
            self.my_graph.remove_edge(replaced_node, tgt, 0)
            self.my_graph.nodes[replaced_node]["is_interior"] = True
            self.my_graph.nodes[replaced_node]["color"] = BiColor.RED
        self.my_graph = nx.compose(other.my_graph, self.my_graph)

        self.multi_edge_permutation.update(other.multi_edge_permutation)
        for replaced_node in relevant_internal:
            bivalent_removed, msg = self.remove_bivalent(replaced_node)
            if not bivalent_removed:
                new_msg = " ".join([f"Got the message {msg} when trying",
                                   f"to remove {replaced_node} as a bivalent vertex.",
                                    "It should have been bivalent because in the constituents",
                                    "it was univalent to distinct targets in the two pieces",
                                    "then the composition would make it bivalent"])
                raise ValueError(new_msg)

        if clear_position:
            self.remove_prop("position")
        # clears everything to do with perfect orientation
        # could fix the perfect orientation instead
        # along this move
        self.__clear_perfect_matching()

        return True, "Success"

    def bdry_to_bdry(self, starting_bdry_node: str) -> Tuple[Optional[BiColor], str]:
        """
        starting at a boundary vertex follow the rules of the road until another
        boundary vertex (possibly the same one)
        the color gives the decorated permutation
        because fixed points are colored by the color of the lollipop
        """
        was_at_this_before = None
        via_edge_keyed = None
        at_this_node = starting_bdry_node
        turn_around_colors: List[BiColor] = []
        path_length = 0
        while True:
            next_node = self.follow_rules_of_road(
                at_this_node, was_at_this_before, via_edge_keyed)
            if next_node is None:
                if at_this_node == starting_bdry_node:
                    if len(turn_around_colors) == 0:
                        return None, at_this_node
                    if path_length == 2:
                        return turn_around_colors[0], at_this_node
                    msg = " ".join(["There was a collapsible tree",
                                    "but couldn't tell what decoration",
                                    "it collapsed to"])
                    raise NotImplementedError(msg)
                return None, at_this_node
            path_length += 1
            if was_at_this_before is not None and next_node[1] == was_at_this_before:
                turn_around_colors.append(self.get_color(at_this_node))
            was_at_this_before = at_this_node
            via_edge_keyed, at_this_node = next_node

    def follow_rules_of_road(self, at_this_node: str,
                             was_at_this_before: Optional[str] = None,
                             via_edge_keyed: Optional[int] = None) -> Optional[Tuple[int, str]]:
        """
        if we are at some vertex and came from some other node, where should we go next
        if we are following the rule that we turn maximally right/left
        depending on the color of where we are and stop if we are on a boundary vertex
        """
        if was_at_this_before is None:
            if not at_this_node in self.all_bdry_nodes:
                raise ValueError(
                    "If don't have a previous vertex, we must be at a boundary vertex")
            return_name, _ = self._by_edge_number(at_this_node, 0)
            return 0, return_name
        if via_edge_keyed is None:
            only_one = self.num_connecting_edges(
                was_at_this_before, at_this_node) == 1
            if only_one:
                key_this_that_all: List[int] =\
                    list(self.my_graph[was_at_this_before]
                         [at_this_node].keys())
                via_edge_keyed = key_this_that_all[0]
            else:
                msg = " ".join(["If there was a previous vertex,",
                                "then there had to have been a",
                                "previous key for the half edge going from",
                                "previous to current"])
                raise ValueError(msg)
        my_color = self.get_color(at_this_node)
        if my_color is None:
            return None
        prev_edge_number = \
            self.__corresponding_multiedge_idx(
                was_at_this_before, at_this_node, via_edge_keyed)
        num_incident_edges = self.my_graph.out_degree(at_this_node)
        if my_color == BiColor.GREEN:
            next_edge_number = prev_edge_number + 1
            if next_edge_number >= num_incident_edges:
                next_edge_number -= num_incident_edges
        else:
            next_edge_number = prev_edge_number - 1
            if next_edge_number < 0:
                next_edge_number += num_incident_edges
        for tgt in self.my_graph[at_this_node]:
            if next_edge_number in self.my_graph[at_this_node][tgt]:
                return next_edge_number, tgt
        raise ValueError(
            f"There should be an edge numbered {next_edge_number} on {at_this_node}")

    def draw(self, *,
             draw_oriented_if_perfect=True,
             show_node_names : bool = True,
             red_nodes: str = "red",
             green_nodes: str = "green",
             bdry_nodes: str = "black",
             oriented_arrows: str = "black",
             unoriented_arrows_perfect: str = "yellow",
             unoriented_arrows_imperfect: str = "black",
             overridden_arrow_orientation : Optional[Callable[[str,str,int],bool]] = None,
             external_circle_color : str = "black",
             internal_circles_color : str = "black",
             show_as_well : bool = True
             ) -> None:
        """
        draw the multigraph without regard to it's planar embedding

        draw_oriented_if_perfect - if the Plabic graph is equipped with a perfect
            orientation, draw the edges as arrows instead of line segments
        show_node_names - show the names of the vertices
        red_nodes,green_nodes,bdry_nodes - the colors of the vertices
            (adjust for black and white printing or red green colorblindness)
        oriented_arrows - what color to draw oriented edges
        unoriented_arrows_perfect,unoriented_arrows_imperfect - if the Plabic graph
            is equipped with a perfect orientation and we are drawing edges without
            orientation then what color to mark the perfect/imperfect edges
        overridden_arrow_orientation - a function which takes the two endpoints
            and key (for multiedges) and if the output is false
            then it doesn't draw that arrowhead for that orientation of the edge
        external_circle_color,internal_circles_color - if the boundaries are on circles
            in a FramedDiskConfig then what color to draw those framed circles
        show_as_well - show right away or let the caller do the plt.show
        """
        color_dict = {BiColor.RED: red_nodes, BiColor.GREEN: green_nodes}

        def name_to_color(name: str) -> str:
            """
            convert from node name to the color to draw it as
            """
            return color_dict.get(
                cast(BiColor, self.get_color(name)), bdry_nodes)

        all_node_names = list(self.my_graph.nodes())
        all_colors = [name_to_color(z) for z in all_node_names]
        if "position" in self.my_extra_props:
            try:
                all_positions = {
                    z: self.my_graph.nodes[z]["position"] for z in all_node_names}
            except KeyError:
                self.remove_prop("position")
                all_positions = None
        else:
            all_positions = None

        something_transparent = "#0f0f0f00"
        if overridden_arrow_orientation is None:
            draw_arrowheads = False
            try:
                special_edge_numbers = {
                    z: self.my_graph.nodes[z]["my_perfect_edge"] for z in all_node_names}
                edge_is_special: List[Tuple[str, str, bool]] = \
                    [(u, v, k == special_edge_numbers[u])
                    for u, v, k in self.my_graph.edges(keys=True)]
                if draw_oriented_if_perfect:
                    draw_arrowheads = True

                    def keep_this_arrow(src_name: str, tgt_name: str, is_special: bool) -> bool:
                        """
                        whether this arrow is kept or not
                        """
                        src_color = self.get_color(src_name)
                        if src_color is None:
                            tgt_color = self.get_color(tgt_name)
                            if tgt_color is None:
                                # boundary to boundary edge
                                # orientation not determined by color
                                # of endpoints
                                return (src_name < tgt_name) ^ is_special
                            return (tgt_color == BiColor.GREEN) ^ (not is_special)
                        return (src_color == BiColor.RED) ^ (not is_special)
                    edge_colors = [oriented_arrows if keep_this_arrow(u, v, is_special)
                                else something_transparent for u, v, is_special in edge_is_special]
                else:
                    edge_colors = [unoriented_arrows_perfect if is_special
                                else unoriented_arrows_imperfect
                                for u, v, is_special in edge_is_special]
            except KeyError:
                edge_colors = None
        else:
            edge_colors = [oriented_arrows if overridden_arrow_orientation(u, v, k)
                        else something_transparent
                        for u, v, k in self.my_graph.edges(keys=True)]
        nx.draw(self.my_graph, pos=all_positions,
                node_color=all_colors, edge_color=edge_colors,
                arrows=draw_arrowheads, with_labels=show_node_names)
        if self.circles_config is not None:
            self.circles_config.draw(on_top_of_existing=True,
                show_as_well=False,
                draw_framing=False,
                external_circle_color=external_circle_color,
                internal_circles_color=internal_circles_color)
        plt.draw()
        if show_as_well:
            plt.show()



"""
the relevant classes for external use 
"""
#pylint:disable=import-error
from .planar_diagram import PlanarNetwork,ChipType,determinant
from .framed_2_disks import FramedDiskConfig
from .plabic_diagram import BiColor,PlabicGraphBuilder,PlabicGraph,ExtraData
from .le_diagram import LeDiagram
from .double_wiring import WiringDiagram
from .triangulation import Triangulation



"""
test Plabic graphs
"""
#pylint:disable=import-error,too-many-locals
from typing import List,Dict
from plabic import PlabicGraph,BiColor,PlabicGraphBuilder

def test_plabic() -> None:
    """
    example from figure 7.1 of https://arxiv.org/pdf/2106.02160.pdf
    black there -> red here
    white there -> green here
    the names of internal vertices are numbered top to bottom
    left to right and noting the one attached to external 6
    is slightly below the 3rd and 4th (though that is hard to see in the picture)
    """
    external_orientation = ["ext1", "ext2", "ext3", "ext4", "ext5", "ext6"]
    _internal_vertices = ["int1", "int2", "int3",
                         "int4", "int5", "int6", "int7", "int8"]
    my_data = {}
    my_data["ext1"] = (BiColor.RED, ["int1"])
    my_data["ext2"] = (BiColor.RED, ["int2"])
    my_data["ext3"] = (BiColor.RED, ["int7"])
    my_data["ext4"] = (BiColor.RED, ["int8"])
    my_data["ext5"] = (BiColor.RED, ["int6"])
    my_data["ext6"] = (BiColor.RED, ["int5"])

    my_data["int1"] = (BiColor.GREEN, ["ext1", "int2", "int3"])
    my_data["int2"] = (BiColor.RED, ["int1", "ext2", "int4"])
    my_data["int3"] = (BiColor.RED, ["int1", "int4", "int6"])
    my_data["int4"] = (BiColor.GREEN, ["int2", "int7", "int3"])
    my_data["int5"] = (BiColor.GREEN, ["ext6"])
    my_data["int6"] = (BiColor.GREEN, ["int3", "int8", "ext5"])
    my_data["int7"] = (BiColor.GREEN, ["int4", "ext3", "int8"])
    my_data["int8"] = (BiColor.RED, ["int6", "int7", "ext4"])
    example_plabic = PlabicGraph(my_data, external_orientation, {})
    assert example_plabic.my_extra_props.issubset([])
    expected_one_steps = ["int1", "int2", "int7", "int8", "int6", "int5"]
    expected_two_steps = ["int2", "int1", "int8", "int7", "int3", "ext6"]
    for cur_bdry_temp, exp_one_step, exp_two_step in\
            zip(external_orientation, expected_one_steps, expected_two_steps):
        one_step = example_plabic.follow_rules_of_road(cur_bdry_temp, None, None)
        assert one_step is not None
        cur_one_step_idx, cur_one_step = one_step
        assert cur_one_step == exp_one_step
        two_step = example_plabic.follow_rules_of_road(
            cur_one_step, cur_bdry_temp, cur_one_step_idx)
        assert two_step is not None
        assert two_step[1] == exp_two_step
    expected_perm = {"ext1":"ext3", "ext2":"ext4",
                         "ext3":"ext5","ext4":"ext1",
                         "ext5":"ext2","ext6":"ext6"}
    expected_decorated = {"ext6":BiColor.GREEN}
    correct_decorated_permutation(example_plabic,external_orientation,
                                  expected_perm,expected_decorated)
    changed, explanation = example_plabic.square_move(("int1", "int2", "int4", "int3"))
    assert changed and explanation == "Success"
    correct_decorated_permutation(example_plabic,external_orientation,
                                  expected_perm,expected_decorated)
    changed, explanation = example_plabic.square_move(("int1", "int2", "int4", "int3"))
    assert changed and explanation == "Success"
    correct_decorated_permutation(example_plabic,external_orientation,
                                  expected_perm,expected_decorated)
    changed, explanation = example_plabic.flip_move("int4", "int7")
    assert changed and explanation == "Success"
    correct_decorated_permutation(example_plabic,external_orientation,
                                  expected_perm,expected_decorated)
    simplifies = example_plabic.greedy_shrink()
    assert simplifies
    simplifies = example_plabic.greedy_shrink()
    assert not simplifies
    correct_decorated_permutation(example_plabic,external_orientation,
                                  expected_perm,expected_decorated)
    did_scale_up = example_plabic.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert not did_scale_up

# pylint:disable = too-many-statements
def test_builder():
    """
    same example as before but with an internal circle
    which has 1 vertex which connects to int5
    """
    my_builder = PlabicGraphBuilder()
    my_builder.set_num_external(6)
    my_builder.add_external_bdry_vertex("ext1",0,"int1")
    my_builder.add_external_bdry_vertex("ext2",1,"int2")
    my_builder.add_external_bdry_vertex("ext3",2,"int7")
    my_builder.add_external_bdry_vertex("ext4",3,"int8")
    my_builder.add_external_bdry_vertex("ext5",4,"int6")
    my_builder.add_external_bdry_vertex("ext6",5,"int5")
    my_builder.set_internal_circles_nums([1])
    my_builder.add_internal_bdry_vertex("bdry1",0,0,"int5")
    my_builder.add_internal_vertex("int1",BiColor.GREEN, ["ext1", "int2", "int3"])
    my_builder.add_internal_vertex("int2",BiColor.RED, ["int1", "ext2", "int4"])
    my_builder.add_internal_vertex("int3",BiColor.RED, ["int1", "int4", "int6"])
    my_builder.add_internal_vertex("int4",BiColor.GREEN, ["int2", "int7", "int3"])
    my_builder.add_internal_vertex("int5",BiColor.GREEN, ["ext6","bdry1"])
    my_builder.add_internal_vertex("int6",BiColor.GREEN, ["int3", "int8", "ext5"])
    my_builder.add_internal_vertex("int7",BiColor.GREEN, ["int4", "ext3", "int8"])
    my_builder.add_internal_vertex("int8",BiColor.RED, ["int6", "int7", "ext4"])
    example_plabic = my_builder.build()
    assert example_plabic.my_extra_props.issubset([])
    exp_external_orientation = ["ext1", "ext2", "ext3", "ext4", "ext5", "ext6"]
    assert example_plabic.my_external_nodes == exp_external_orientation
    expected_one_steps = ["int1", "int2", "int7", "int8", "int6", "int5"]
    expected_two_steps = ["int2", "int1", "int8", "int7", "int3", "bdry1"]
    for cur_bdry_temp, exp_one_step, exp_two_step in\
            zip(exp_external_orientation, expected_one_steps, expected_two_steps):
        one_step = example_plabic.follow_rules_of_road(cur_bdry_temp, None, None)
        assert one_step is not None
        cur_one_step_idx, cur_one_step = one_step
        assert cur_one_step == exp_one_step
        two_step = example_plabic.follow_rules_of_road(
            cur_one_step, cur_bdry_temp, cur_one_step_idx)
        assert two_step is not None
        assert two_step[1] == exp_two_step
    expected_perm = {"ext1":"ext3", "ext2":"ext4",
                         "ext3":"ext5","ext4":"ext1",
                         "ext5":"ext2","ext6":"bdry1","bdry1":"ext6"}
    expected_decorated = {}
    correct_decorated_permutation(example_plabic,exp_external_orientation,
                                  expected_perm,expected_decorated)
    changed, explanation = example_plabic.square_move(("int1", "int2", "int4", "int3"))
    assert changed and explanation == "Success"
    correct_decorated_permutation(example_plabic,exp_external_orientation,
                                  expected_perm,expected_decorated)
    changed, explanation = example_plabic.square_move(("int1", "int2", "int4", "int3"))
    assert changed and explanation == "Success"
    correct_decorated_permutation(example_plabic,exp_external_orientation,
                                  expected_perm,expected_decorated)
    changed, explanation = example_plabic.flip_move("int4", "int7")
    assert changed and explanation == "Success"
    correct_decorated_permutation(example_plabic,exp_external_orientation,
                                  expected_perm,expected_decorated)
    changed, explanation = example_plabic.remove_bivalent("int5")
    assert not changed and explanation == "Cannot cause two boundary nodes to connect directly"
    simplifies = example_plabic.greedy_shrink()
    assert simplifies
    simplifies = example_plabic.greedy_shrink()
    assert not simplifies
    correct_decorated_permutation(example_plabic,exp_external_orientation,
                                  expected_perm,expected_decorated)
    did_scale_up = example_plabic.coordinate_transform(lambda z: (z[0]*2,z[1]*2))
    assert not did_scale_up

def correct_decorated_permutation(example_plabic : PlabicGraph,
                                        external_orientation : List[str],
                                        expected_perm : Dict[str,str],
                                        expected_decorated : Dict[str,BiColor]):
    """
    assert that the decorated permutation is as expected
    """
    for cur_bdry_temp in external_orientation:
        turn_color, cur_tgt = example_plabic.bdry_to_bdry(cur_bdry_temp)
        assert cur_tgt == expected_perm[cur_bdry_temp]
        assert turn_color == expected_decorated.get(cur_bdry_temp,None)
